# -*- coding: utf-8 -*-

"""Buffer for dealing with out-of-order/missing frames."""

import copy

__version__ = '0.3'
__revision__ = '$ Revision: 5 $'
__all__ = ['FrameBuffer', 'TBNFrameBuffer', 'DRXFrameBuffer', 'TBWFrameBuffer', '__version__', '__revision__', '__all__']


def cmpStands(x, y):
	"""Function to compare two frames and sort by stand/beam number."""
	
	# Parse if frame IDs to extract the stand/beam, tunning, and polarization
	# information (where appropriate)
	idsX = x.parseID()
	idsY = y.parseID()
	try:
		len1 = len(idsX)
		if len1 == 2:
			sX = 2*(idsX[0]-1) + idsX[1]
			sY = 2*(idsY[0]-1) + idsY[1]
		else:
			sX = 4*(idsX[0]-1) + 2*(idsX[1]-1) + idsX[2]
			sY = 4*(idsY[0]-1) + 2*(idsY[1]-1) + idsY[2]
	except TypeError:
		sX = idsX
		sY = idsY
		
	# Do the comparison
	if sY > sX:
		return -1
	elif sX > sY:
		return 1
	else:
		return 0
	

class FrameBuffer(object):
	"""Frame buffer for re-ordering TBN and DRX frames in time order.  
	This class is filled with frames and a returns a frame list when 
	either:
	  * the buffer contains all frames for a given time, or
	  * another has started to be filled and has more than 50 frames in 
	    it.

	The buffer also keeps track of what has already been read out so 
	that tardy frames are just dropped.  For buffers that are read out,
	missing frames are replaced with frames filled with zeros."""

	def __init__(self, stands=[], pols=[], beams=[], tunes=[]):
		"""Initialize the buffer with a list of:
		  * TBN
		      * list of stands
		      * list of pols
		  * DRX
		      * list of beams
		      * list of tunnings
		      * list of pols
		By doing this, we should be able to keep up with when the buffer 
		is full and to help figure out which stands are missing."""

		# The buffer itself
		self.buffer = {}
		self.done = []
		
		# Buffer statistics
		self.full = 0		# Number of times a full buffer was emptied
		self.partial = 0	# Number of times a partial buffer was emptied
		self.dropped = 0	# Number of late frames dropped
		self.missing = 0	# Number of missing frames
		
		# Information describing the obserations
		self.stands = stands
		self.pols = pols
		self.beams = beams
		self.tunes = tunes
		
		# Figure out how many frames fill the buffer and the list of all
		# possible frames in the data set
		calcFramesOut = self.__calcFrames()
		self.nFrames = calcFramesOut[0]
		self.possibleFrames = calcFramesOut[1]
		
		# When to dump the dump partial packets
		self.dump = 50
		
	def __calcFrames(self):
		"""Calculate the maximum number of frames that we expect from 
		the setup of the observations and a list of tuples that describe
		all of the possible stand/beam/pol/tune combination."""
		
		# TBN Mode
		if len(self.beams) == 0:
			nFrames = len(self.stands)*len(self.pols)
				
			frameList = []
			for stand in self.stands:
				for pol in self.pols:
					frameList.append((stand,pol))
				
		# DRX mode
		else:
			nFrames = len(self.beams)*len(self.tunes)*len(self.pols)
			
			frameList = []
			for beam in self.beams:
				for tune in self.tunes:
					for pol in self.pols:
						frameList.append((beam, tune, pol))
			
		return (nFrames, frameList)

	def append(self, frame):
		"""Append a new frame to the buffer with the appropriate time
		tag.  True is returned if the frame was added to the buffer and
		False if the frame was dropped."""

		# Make sure that it is not in the `done' list.  If it is,
		# disgaurd the frame and make a note of it.
		if frame.data.timeTag in self.done:
			self.dropped = self.dropped + 1
			return False

		# If that time tag hasn't been done yet, add it to the 
		# buffer in the correct place.
		if frame.data.timeTag not in self.buffer.keys():
			self.buffer[frame.data.timeTag] = []
		self.buffer[frame.data.timeTag].append(frame)
		return True

	def get(self):
		"""Return a list of frames that consitute a 'full' buffer.  
		Afterwards, delete that buffer."""

		# Get the current status of the buffer
		nKeys, minKey, minCount, maxKey, maxCount = self.__bufferStatus()
		if nKeys == 0:
			output = None
		elif nKeys == 1:
			if maxCount == self.nFrames:
				self.done.append(maxKey)
				self.full = self.full + 1
				
				output = copy.deepcopy(self.buffer[maxKey])
				del(self.buffer[maxKey])
			else:
				output = None
		else:
			if maxCount == self.nFrames:
				self.done.append(maxKey)
				self.full = self.full + 1
				
				output = copy.deepcopy(self.buffer[maxKey])
				del(self.buffer[maxKey])
			elif minCount == self.dump:
				self.partial = self.partial + 1
				self.missing = self.missing + (len(self.buffer[maxKey]) - self.nFrames)
				self.done.append(maxKey)
				
				output = copy.deepcopy(self.buffer[maxKey])
				for frame in self.__missingList(maxKey):
					output.append( self.__createFrame(key, frame) )
				
				del(self.buffer[maxKey])
				print "%i is missing Frames" % maxKey
			else:
				output = None
		
		# Sort and return
		if output is None:
			return output
		else:
			return sorted(output, cmp=cmpStands)

	def __bufferStatus(self):
		"""Return the state of the buffer in the form of a tuple.  
		The values are:
		  * number of buffer keys, 
		  * buffer key with minimum length, 
		  * length of the shortest buffer, 
		  * buffer key with maximum length, and
		  * length of the longest buffer.
		If the buffer is empty, min/maxKeys are set to None."""

		# Build the list of keys to look at
		keyList = self.buffer.keys()

		# Setup the basic variables
		nKeys = len(keyList)
		minKey = None
		minCount = self.nFrames+1
		maxKey = None
		maxCount = 0
		
		# Loop to fill in the min/max values
		keyList = self.buffer.keys()
		for key, count in zip(keyList, [len(self.buffer[key]) for key in keyList]):
			if count < minCount:
				minKey = key
				minCount = count
			if count > maxCount:
				maxKey = key
				maxCount = count
				
		return (nKeys, minKey, minCount, maxKey, maxCount)

	def __missingList(self, key):
		"""Create a list of tuples of missing frame information."""
		
		# Find out what frames we have
		frameList = []
		for frame in self.buffer[key]:
			frameList.append(frame.parseID())
			
		# Compare the existing list with the possible list stored in the 
		# FrameBuffer object to build a list of missing frames.
		missingList = []
		for frame in self.frameList:
			if frame not in frameList:
				missingList.append(frame)
				
		return missingList

	def __createFill(self, key, frameParameters):
		"""Create a 'fill' frame of zeros using an existing good
		packet as a template."""

		fillFrame = copy.deepcopy(self.buffer[key][0])
		
		# Get out the frame parameters
		if len(frameParameters) == 2:
			stand, pol = frameParameters
		else:
			bean, tune, pol = frameParameters
		
		# Fix-up the header
		try:
			fillFrame.header.tbnID = 2*(stand-1) + pol + 1
		except AttributeError:
			pass
		try:
			fillFrame.header.drxID = (beam & 7) | ((tune & 7) << 3) | ((pol & 1) << 7)
		except AttributeError:
			pass
		fillFrame.header.frameCount = 0
		
		# Zero the data for the fill packet
		fillFrame.data.iq *= 0
		
		# Invalidate the frame
		fillFrame.valid = False
		
		return fillFrame
		
	def status(self):
		"""Print out the status of the buffer."""
		
		nf = 0
		for key in self.buffer.keys():
			nf = nf + len(self.buffer[key])
			
		outString = ''
		outString = '\n'.join([outString, "Current buffer level:  %i frames" % nf])
		outString = '\n'.join([outString, "Buffer dumps:  %i full / %i partial" % (self.filled, self.partial)])
		outString = '\n'.join([outString, "--"])
		outString = '\n'.join([outString, "Missing frames:  %i" % self.missing])
		outString = '\n'.join([outString, "Dropped frames:  %i" % self.dropped])

		print outString


class TBNFrameBuffer(FrameBuffer):
	"""A sub-type of FrameBuffer specifically for dealing with TBN frames.
	See FrameBuffer for a description of how the buffering is implemented."""
	
	def __init__(self, stands=[], pols=[0, 1]):
		super(TBNFrameBuffer, self).__init__(stands=stands, pols=pols)


class DRXFrameBuffer(FrameBuffer):
	"""A sub-type of FrameBuffer specifically for dealing with DRX frames.
	See FrameBuffer for a description of how the buffering is implemented."""
	
	def __init__(self, beams=[], tunes=[1, 2], pols=[1, 2]):
		super(DRXFrameBuffer, self).__init__(beams=beams, tunes=tunes, pols=pols)


class TBWFrameBuffer(object):
	"""Frame buffer for re-ordering TBW frames in time order.  
	This class is filled with frames and a returns a frame list when 
	either:
	  * the buffer contains all frames for a given time, or
	  * another has started to be filled and has more than 5 frames in 
	    it.

	The buffer also keeps track of what has already been read out so 
	that tardy frames are just dropped.  For buffers that are read out,
	missing frames are replaced with frames filled with zeros.
	
	TBW cannot be treated in the same way as TBN and DRX because of how 
	the packets are sent out by DP, i.e., in stands pairs rather than by 
	time."""

	def __init__(self, stands=[], nFrames=None):
		"""Initialize the buffer with a list of:
		  * TBW:
		      * list of stands
		By doing this, we should be able to keep up with when the buffer 
		is full and to help figure out which stands are missing."""

		# The buffer itself
		self.buffer = {}
		self.done = []
		
		# Buffer statistics
		self.full = 0		# Number of times a full buffer was emptied
		self.partial = 0	# Number of times a partial buffer was emptied
		self.dropped = 0	# Number of late frames dropped
		self.missing = 0	# Number of missing frames
		
		# Information describing the obserations
		self.stands = stands
		
		# Figure out how many frames fill the buffer and the list of all
		# possible frames in the data set
		self.nFrames = nFrames
		self.possibleFrames = self.stands
		
		# When to dump the dump partial packets
		self.dump = 50

	def append(self, frame):
		"""Append a new frame to the buffer with the appropriate time
		tag.  True is returned if the frame was added to the buffer and
		False if the frame was dropped."""

		# Make sure that it is not in the `done' list.  If it is,
		# disgaurd the frame and make a note of it.
		if frame.parseID() in self.done:
			self.dropped = self.dropped + 1
			return False

		# If that time tag hasn't been done yet, add it to the 
		# buffer in the correct place.
		if frame.parseID() not in self.buffer.keys():
			self.buffer[frame.parseID()] = []
		self.buffer[frame.parseID()].append(frame)
		return True

	def get(self):
		"""Return a list of frames that consitute a 'full' buffer.  
		Afterwards, delete that buffer."""
		
		# Get the current status of the buffer
		nKeys, minKey, minCount, maxKey, maxCount = self.__bufferStatus()
		if nKeys <= 1:
			output = None
		elif nKeys == 2:
			if maxCount == self.nFrames:
				self.done.append(maxKey)
				self.full = self.full + 1
				
				output = copy.deepcopy(self.buffer[maxKey])
				del(self.buffer[maxKey])
			else:
				output = None
		else:
			if maxCount == self.nFrames:
				self.done.append(maxKey)
				self.full = self.full + 1
				
				output = copy.deepcopy(self.buffer[maxKey])
				del(self.buffer[maxKey])
			elif minCount == self.dump:
				self.partial = self.partial + 1
				self.missing = self.missing + (len(self.buffer[maxKey]) - self.nFrames)
				self.done.append(maxKey)
				
				output = copy.deepcopy(self.buffer[maxKey])
				del(self.buffer[maxKey])
				print "%i is missing Frames" % maxKey
			else:
				output = None
		# Sort and return
		if output is None:
			return output
		else:
			return sorted(output)

	def __bufferStatus(self):
		"""Return the state of the buffer in the form of a tuple.  
		The values are:
		  * number of buffer keys, 
		  * buffer key with minimum length, 
		  * length of the shortest buffer, 
		  * buffer key with maximum length, and
		  * length of the longest buffer.
		If the buffer is empty, min/maxKeys are set to None."""

		# Build the list of keys to look at

		keyList = self.buffer.keys()
		# Setup the basic variables
		nKeys = len(keyList)
		minKey = None
		minCount = self.nFrames+1
		maxKey = None
		maxCount = 0
		
		# Loop to fill in the min/max values
		keyList = self.buffer.keys()
		for key, count in zip(keyList, [len(self.buffer[key]) for key in keyList]):
			if count < minCount:
				minKey = key
				minCount = count
			if count > maxCount:
				maxKey = key
				maxCount = count
				
		return (nKeys, minKey, minCount, maxKey, maxCount)
		
	def status(self):
		"""Print out the status of the buffer."""
		
		nf = 0
		for key in self.buffer.keys():
			nf = nf + len(self.buffer[key])
			
		outString = ''
		outString = '\n'.join([outString, "Current buffer level:  %i frames" % nf])
		outString = '\n'.join([outString, "Buffer dumps:  %i full / %i partial" % (self.filled, self.partial)])
		outString = '\n'.join([outString, "--"])
		outString = '\n'.join([outString, "Missing frames:  %i" % self.missing])
		outString = '\n'.join([outString, "Dropped frames:  %i" % self.dropped])

		print outString


	# Stuff that may work someday.  The trick is figuring out which 
	# time stamps are missing from a sequence:
	"""
	From get:
		for frame in self.__missingList(maxKey):
			output.append( self.__createFrame(key, frame) )
	
	def __missingList(self, key):
		Create a list of tuples of missing frame information.
		
		frameList = []
		for frame in self.buffer[key]:
			frameList.append(frame.parseID())
			

	def __createFill(self, key, frameParameters):
		Create a 'fill' frame of zeros using an existing good
		packet as a template.

		fillFrame = copy.deepcopy(self.buffer[key][0])
		
		# Fix-up the header
		fillFrame.header.frameCount = 0
		
		# Zero the data for the fill packet
		fillFrame.data.timeTag = frameParameters
		fillFrame.data.xy *= 0
		
		# Invalidate the frame
		fillFrame.valid = False
		
		return fillFrame
	"""


# -*- coding: utf-8 -*-

"""Module that contains the warning classes for the DRX, TBN, and TBW 
readers.  The two warnings defined here are:
  * warnDeprecated and
  * warnExperimental.

These are for warning users about deprecated functions and experimental 
features, respectively."""

import sys
import warnings as sysWarnings
import linecache

__version__ = '0.2'
__revision__ = '$ Revision: 4 $'
__all__ = ['warnDeprecated', 'warnExperimental', '__version__', '__revision__', '__all__']


def __modifiedFormatWarning(message, category, filename, lineno, line=None):
	"""For some reason, the default warning.formatwarning function does not 
	support the 'line' argument and all warning calls throw their own warning
	calls.  This function is meant to fix that."""

	s =  "%s:%s: %s: %s\n" % (filename, lineno, category.__name__, message)
	if line is None:
		line = linecache.getline(filename, lineno)
	if line:
		line = line.strip()
		s += "  %s\n" % line
	return s


def __modifiedShowWarning(message, category, filename, lineno, file=None, line=None):
	"""This is a new warning.showwarning function to address the same problem as 
	seen with the warning.formatwarning default function."""

	s =  __modifiedFormatWarning(message, category, filename, lineno, line=line)
	
	if file is None:
		sys.stderr.write(s)
	else:
		file.write(s)


def warnDeprecated(name, memo=None):
	"""Wrapper around warnins.warn to fix the problems with 
	[(format)(show)]warnings functions described above."""

	sysWarnings.showwarning = __modifiedShowWarning
	sysWarnings.formatwarning = __modifiedFormatWarning
	
	if memo is None:
		sysWarnings.warn("'%s' is deprecated" % name, DeprecationWarning, 3)
	else:
		sysWarnings.warn("'%s' is deprecated: %s" % (name, memo), DeprecationWarning, 3)


def warnExperimental(name, memo=None):
	"""Wrapper around warnins.warn to fix the problems with 
	[(format)(show)]warnings functions described above."""

	sysWarnings.showwarning = __modifiedShowWarning
	sysWarnings.formatwarning = __modifiedFormatWarning
	
	if memo is None:
		sysWarnings.warn("'%s' is experimental" % name, ImportWarning, 3)
	else:
		sysWarnings.warn("'%s' is experimental: %s" % (name, memo), ImportWarning, 3)



# -*- coding: utf-8 -*-

"""Module to simulate observations made with the DP system."""

import functools
import itertools
import aipy
import math
import time
import numpy
import scipy.stats

from lsl import astro
from lsl.common import dp as dp_common
from lsl.common import stations as lwa_common
from lsl.sim import tbw
from lsl.sim import tbn
from lsl.sim import drx
from lsl.sim import vis
from lsl.reader.tbn import filterCodes as TBNFilters
from lsl.reader.drx import filterCodes as DRXFilters

__version__ = '0.1'
__revision__ = '$ Revision: 1 $'
__all__ = ['basicSignal', '__version__', '__revision__', '__all__']


def basicSignal(fh, stands, nFrames, mode='DRX', filter=6, bits=12, tStart=0):
	"""Generate a collection of frames with a basic test signal for TBW, TBN, 
	and DRX.  The signals for the three modes are:
	
	TBW
	    * noise + 40 MHz signal for x-pol.
	    * noise + 60 MHz signal for y-pol.
	      -> odd stands
	    * noise + 30 MHz signal for x-pol.
	    * noise + 50 MHz signal for ypol.
	      -> even stands
	
	TBN
	    * noise + (sampleRate/4) kHz signal for x-pol. and noise + 
	      (-sampleRate/4) for y-pol.

	DRX
	    * same test signal used in the original lwa_dp_sim
	    
	All modes need to have stands (beams in the case of DRX) and number of
	frames to generate.  TBW also needs to 'bits' keyword set to generate 
	either 12-bit or 4-bit data.  The TBN and DRX frames need the 'filter'
	keyword set to specify the filter width.  In addition, the 'stands' 
	argument is interpreted as beam numbers for DRX."""

	if tStart == 0:
		tStart = time.time()
	else:
		tStart = float(tStart)

	if mode == 'TBW':
		sampleRate = dp_common.fS
		
		if bits == 12:
			maxValue = 2047
			samplesPerFrame = 400
		else:
			maxValue =  7
			samplesPerFrame = 1200

		nCaptures = int(math.ceil(nFrames / 30000.0))
		for capture in range(nCaptures):
			for stand1, stand2 in zip(stands[0::2], stands[1::2]):
				print "Simulating TBW capture %i, stands %i and %i" % (capture+1, stand1, stand2)
				FramesThisBatch = nFrames - capture*30000
				if FramesThisBatch > 30000:
					FramesThisBatch = 30000
				print "-> Frames %i" % FramesThisBatch
				for i in range(FramesThisBatch):
					t = tStart + i*samplesPerFrame/sampleRate + 60.0*capture
					
					cFrame = tbw.SimFrame(stand=stand1, frameCount=i+1, dataBits=bits, obsTime=t)
					cFrame.xy = numpy.random.randn(2, samplesPerFrame)
					cFrame.xy[0,:] *= maxValue/15.0
					cFrame.xy[0,:] += maxValue*numpy.cos(2*numpy.pi*0.2041*numpy.arange(samplesPerFrame))
					cFrame.xy[1,:] *= maxValue/15.0
					cFrame.xy[1,:] += maxValue*numpy.cos(2*numpy.pi*0.3061*numpy.arange(samplesPerFrame))
					
					cFrame.writeRawFrame(fh)

					cFrame = tbw.SimFrame(stand=stand2, frameCount=i+1, dataBits=bits, obsTime=t)
					cFrame.xy = numpy.random.randn(2, samplesPerFrame)
					cFrame.xy[0,:] *= maxValue/15.0
					cFrame.xy[0,:] += maxValue*numpy.cos(2*numpy.pi*0.1531*numpy.arange(samplesPerFrame))
					cFrame.xy[1,:] *= maxValue/15.0
					cFrame.xy[1,:] += maxValue*numpy.cos(2*numpy.pi*0.2551*numpy.arange(samplesPerFrame))
					
					cFrame.writeRawFrame(fh)
	
	if mode == 'TBN':
		sampleRate = TBNFilters[filter]
		maxValue = 127
		samplesPerFrame = 512
		upperSpike = sampleRate / 4.0
		lowerSpike = -sampleRate / 4.0
		
		for i in range(nFrames):
			if i % 1000 == 0:
				print "Simulating TBN frame %i" % (i+1)
			t = tStart + i*samplesPerFrame/sampleRate
			for stand in stands:
				cFrame = tbn.SimFrame(stand=stand, pol=0, frameCount=i+1, obsTime=t)
				temp = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
				temp.real = numpy.random.randn(samplesPerFrame)
				temp.real *= maxValue/15.0
				temp.real += maxValue*numpy.cos(2*numpy.pi*upperSpike*(t+numpy.arange(samplesPerFrame)/sampleRate))
				temp.imag = numpy.random.randn(samplesPerFrame)
				temp.imag *= maxValue/15.0
				temp.imag += maxValue*numpy.sin(2*numpy.pi*upperSpike*(t+numpy.arange(samplesPerFrame)/sampleRate))
				cFrame.iq = temp
				
				cFrame.writeRawFrame(fh)

				cFrame = tbn.SimFrame(stand=stand, pol=1, frameCount=i+1, obsTime=t)
				temp = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
				temp.real = numpy.random.randn(samplesPerFrame)
				temp.real *= maxValue/15.0
				temp.real += maxValue*numpy.cos(2*numpy.pi*lowerSpike*(t+numpy.arange(samplesPerFrame)/sampleRate))
				temp.imag = numpy.random.randn(samplesPerFrame)
				temp.imag *= maxValue/15.0
				temp.imag += maxValue*numpy.sin(2*numpy.pi*lowerSpike*(t+numpy.arange(samplesPerFrame)/sampleRate))
				cFrame.iq = temp
				
				cFrame.writeRawFrame(fh)
	
	if mode == 'DRX':
		sampleRate = DRXFilters[filter]
		maxValue = 7
		samplesPerFrame = 4096
		beams = stands
		
		part1 = scipy.stats.norm(loc=1000, scale=100, size=samplesPerFrame)
		part2 = scipy.stats.norm(loc=3000, scale=100, size=samplesPerFrame)
		part3 = scipy.stats.norm(loc=1500, scale=100, size=samplesPerFrame)
		part4 = scipy.stats.norm(loc=2500, scale=100, size=samplesPerFrame)

		signal1 = part1.pdf(numpy.arange(samplesPerFrame)) + 0.3*part2.pdf(numpy.arange(samplesPerFrame))
		signal1 *= 0.2 / signal1.max() 
		signal2 = 0.3*part1.pdf(numpy.arange(samplesPerFrame)) + part2.pdf(numpy.arange(samplesPerFrame))
		signal2 *= 0.2 / signal2.max() 
		signal3 = part3.pdf(numpy.arange(samplesPerFrame)) + 0.3*part4.pdf(numpy.arange(samplesPerFrame))
		signal3 *= 0.2 / signal3.max() 
		signal4 = 0.3*part3.pdf(numpy.arange(samplesPerFrame)) + part4.pdf(numpy.arange(samplesPerFrame))
		signal4 *= 0.2 / signal4.max() 
		signal = [signal1, signal2, signal3, signal4]
	
		norm = functools.partial(numpy.random.normal, loc=8)

		for i in range(nFrames):
			if i % 1000 == 0:
				print "Simulating DRX frame %i" % i
			t = tStart + i*samplesPerFrame/sampleRate
			for beam in beams:
				for tune in [1, 2]:
					for pol in [0, 1]:
						cFrame = drx.SimFrame(beam=beam, tune=tune, pol=pol, frameCount=i+1, filterCode=filter, timeOffset=0, obsTime=t, flags=0)

						iq = numpy.zeros(samplesPerFrame, dtype=numpy.singlecomplex)
						for chan, scale in itertools.izip(numpy.arange(samplesPerFrame), signal[2*(tune-1)+pol]):
							iq.real[chan] = norm(scale=scale)
							iq.imag[chan] = norm(scale=scale)
						cFrame.iq = iq

						cFrame.writeRawFrame(fh)


def __getAntennaArray(station, stands, time, freqs):
	"""Given a LWA station object, a list of stands, an observation time, and
	a list of frequencies in Hz, build an aipy AntennaArray object."""

	return vis.buildSimArray(station, stands, freqs/1e9, jd=astro.unix_to_utcjd(time))


def __getSourceParameters(aa, time, srcs):
	"""Given an aipy AntennaArray object, an observation time, and aipy.src 
	object, return all of the parameters needed for a simulation."""
	
	# Set the time for the array
	aa.set_unixtime(time)

	# Compute the source parameters
	srcs_tp = []
	srcs_mp = []
	srcs_jy = []
	srcs_fq = []
	for name,src in srcs.iteritems():
		## Update the source's coordinates
		src.compute(aa)

		## Get parameters
		top = src.get_crds(crdsys='top', ncrd=3)	# topo. coords.
		map = src.map							# equitorial -> topo. rotation matrix
		jys = src.get_jys()						# F_nu
		frq = aa.get_afreqs()					# nu

		## Fix the lowest frequencies to avoid problems with the flux blowing up
		## at nu = 0 Hz by replacing flux values below 1 MHz with the flux at 
		## 1 MHz
		Jyat1MHz = jys[ numpy.where( numpy.abs(frq-0.001) == numpy.abs(frq-0.001).min() ) ]
		jys = numpy.where( frq >= 0.001, jys, Jyat1MHz )

		## Filter out sources that are below the horizon or have no flux
		srcAzAlt = aipy.coord.top2azalt(top) * 180/math.pi
		if srcAzAlt[1] <= 0 or jys.sum() <= 0:
			continue

		## Save values into the source arrays
		srcs_tp.append( top )
		srcs_mp.append( map )
		srcs_jy.append( jys )
		srcs_fq.append( frq )

	# Return the values as a dictionary
	return {'topo': srcs_tp, 'trans': srcs_mp, 'flux': srcs_jy, 'freq': srcs_fq}


def __buildSignals(aa, srcParams, times):
	"""Given an aipy AntennaArray, a list of stand numbers, a dictionary of source parameters, and an 
	array of times in ns, return a numpy array of the simulated signals that is 
	Nstands x Ntimes in shape."""

	# Find out how many stands, srcs, and samples (times) we are working with
	Nstand = len(aa.ants)
	Nsrc = len(srcParams['topo'])
	Ntime = len(times)

	# Define the stand position and cable/signal delay caches to the simulator 
	# move along faster
	dlyCache = uvUtils.SignalCache(aa.get_afreqs()*1e9, applyDispersion=True)

	# Setup a temporary array to hold the signals per source, stand, and time.  
	# This array is complex so that it can accomidate both TBW and TBN data at
	# the same time
	temp = n.zeros((Nsrc, Nstand, Ntime), dtype=n.complex64)

	# Loop over sources and stands to build up the signals
	srcCount = 0
	for topo,trans,flux,freq in zip(srcParams['topo'], srcParams['trans'], srcParams['flux'], srcParams['freq']):
		antCount = 0
		for ant in aa.ants:
			# Zeroth, get the beam response in the direction of the current source for all frequencies
			antResponse = numpy.squeeze( ant.bm_response(tp, pol=pol) )

			# First, do the geometric delay
			geoDelay = ( numpy.dot(trans, ant.pos).transpose() )[2] 

			# Second, do the cable delay
			Delayat1MHz = dlyCache.cableDelay(ant.stand, freq=1.0e6)[0] * 1e9 # s -> ns
			cblDelay = dlyCache.cableDelay(ant.stand) * 1e9 # s -> ns
			# NB: Replace the cable delays below 1 MHz with the 1 MHz value to keep the 
			# delays from blowing up for small f
			cblDelay = numpy.where( freq >= 0.001, cblDelay, Delayat1MHz )

			for a,j,f,d in zip(antResponse, flux, freq, cblDelay):
				factor = a * n.sqrt(j)
				angle = 2*n.pi*f*(t + (d - geoDelay))
				temp[srcCount,antCount,:] += factor*(n.cos(angle) + 1j*n.sin(angle))
			antCount = antCount + 1
		srcCount = srcCount + 1

	# Sum over sources and done
	tdSignals = temp.sum(axis=0)
	return tdSignals


def pointSource(fh, stands, src, nFrames, mode='TBN', filter=7, bits=12, tStart=0):
	"""Generate a collection of frames with a point source signal for TBW
	and TBN.  The point source is specified as a aipy.src object.
	    
	All modes need to have stands (beams in the case of DRX) and number of
	frames to generate.  TBW also needs to `bits' keyword set to generate 
	either 12-bit or 4-bit data.  The TBN frames need the `filter' keyword 
	set to specify the filter width."""

	if tStart == 0:
		tStart = time.time()
	else:
		tStart = float(tStart)

	if mode == 'TBW':
		sampleRate = dp_common.fS
		freqs = (numpy.fft.fftfreq(1024, d=1.0/sampleRate))[1:512]
		aa = __buildAntennaArray(lwa_common.lwa1(), stands, tStart, freqs)
		
		if bits == 12:
			maxValue = 2047
			samplesPerFrame = 400
		else:
			maxValue =  7
			samplesPerFrame = 1200

		nCaptures = int(math.ceil(nFrames / 30000.0))
		for capture in range(nCaptures):
			j = 0
			k = 1
			for stand1, stand2 in zip(stands[0::2], stands[1::2]):
				print "Simulating TBW capture %i, stands %i and %i" % (capture+1, stand1, stand2)
				FramesThisBatch = nFrames - capture*30000
				if FramesThisBatch > 30000:
					FramesThisBatch = 30000
				print "-> Frames %i" % FramesThisBatch

				for i in range(FramesThisBatch):
					frameT = tStart + i*samplesPerFrame/sampleRate + 60.0*capture

					# Get the source parameters
					srcParams = __getSourceParameters(aa, frameT, src)

					# Generate the time series response of each signal at each frequency
					t = frameT + n.arange(samplesPerFrame)/sampleRate * 1e9 # s -> ns
					tdSignals = __buildSignals(aa, srcParams, t)

					cFrame = tbw.SimFrame(stand=stand1, frameCount=i+1, dataBits=bits, obsTime=frameT)
					cFrame.xy = numpy.random.randn(2, samplesPerFrame)
					cFrame.xy[0,:] = tdSignals.real[j,:]
					cFrame.xy[1,:] = tdSignals.real[j,:]
					
					cFrame.writeRawFrame(fh)

					cFrame = tbw.SimFrame(stand=stand2, frameCount=i+1, dataBits=bits, obsTime=frameT)
					cFrame.xy = numpy.random.randn(2, samplesPerFrame)
					cFrame.xy[0,:] = tdSignals.real[k,:]
					cFrame.xy[1,:] = tdSignals.real[k,:]
					
					cFrame.writeRawFrame(fh)
			j = j + 2
			k = k + 2

	if mode == 'TBN':
		sampleRate = TBNFilters[filter]
		maxValue = 127
		samplesPerFrame = 512
		freqs = (numpy.fft.fftfreq(samplesPerFrame, d=1.0/sampleRate)) + CentralFreq
		freqs = numpy.fft.fftshift(freqs)
		aa = __buildAntennaArray(lwa_common.lwa1(), stands, tStart, freqs)

		for i in range(nFrames):
			if i % 1000 == 0:
				print "Simulating TBN frame %i" % (i+1)
			frameT = tStart + i*samplesPerFrame/sampleRate

			# Get the source parameters
			srcParams = __getSourceParameters(aa, frameT, src)

			# Generate the time series response of each signal at each frequency
			t = frameT + n.arange(samplesPerFrame)/sampleRate * 1e9 # s -> ns
			tdSignals = __buildSignals(aa, srcParams, t)

			j = 0
			for stand in stands:
				cFrame = tbn.SimFrame(stand=stand, pol=0, frameCount=i+1, obsTime=frameT)
				cFrame.iq = tdSignals[j,:].astype(numpy.singlecomplex)
				
				cFrame.writeRawFrame(fh)

				cFrame = tbn.SimFrame(stand=stand, pol=1, frameCount=i+1, obsTime=frameT)
				cFrame.iq = tdSignals[j,:].astype(numpy.singlecomplex)
				
				cFrame.writeRawFrame(fh)
				
				j = j + 1



# -*- coding: utf-8 -*-

"""Python module to read in DRX data.  This module defines the following 
classes for storing the DRX data found in a file:

Frame
  object that contains all data associated with a particular DRX frame.  
  The primary constituents of each frame are:
    * FrameHeader - the DRX frame header object and
    * FrameData   - the DRX frame data object.
  Combined, these two objects contain all of the information found in the 
  original DRX frame.

ObservingBlock
  object that stores a collection of Frames for all beams/tunnings/
  polarizations for a particular time.

The functions defined in this module fall into two class:
 1. convert a frame in a file to a Frame object and
 2. describe the format of the data in the file.

For reading in data, use the readFrame function.  It takes a python file-
handle as an input and returns a fully-filled Frame object.  The readBlock
function reads in a (user-defined) number of DRX frames and returns a 
ObservingBlock object.

For describing the format of data in the file, two function are provided:

getBeamCount
  read in the first few frames of an open file handle and return how many 
  beams are present in the file.

getFramesPerObs
  read in the first several frames to see how many frames (tunnings/polarizations)
  are associated with each beam.

Finally, there are also two experimental functions defined in this file for 
averaging DRX observations.  averageObservations performs a straight channel-
by-channel average of a collection of DRX frames, which averageObservations2
allows for both temporal and spatial averaging.
"""

import os
import sys
import math
import time
import numpy
import struct

from  lsl.common import dp as dp_common
from errors import *

__version__ = '0.4'
__revision__ = '$ Revision: 16 $'
__all__ = ['FrameHeader', 'FrameData', 'Frame', 'ObservingBlock', 'readFrame', 'readBlock', 'getBeamCount', 'getFramesPerObs', 'averageObservations', 'averageObservations2', 'FrameSize', 'filterCodes', '__version__', '__revision__', '__all__']

FrameSize = 4128

# List of filter codes and their corresponding sample rates in Hz
filterCodes = {1: 250000, 2: 500000, 3: 1000000, 4: 2000000, 5: 4000000, 6: 9800000, 7: 19600000}


class FrameHeader(object):
	"""Class that stores the information found in the header of a DRX 
	frame.  All six fields listed in the DP IDC version H are stored as 
	well as the original binary header data."""
	
	def __init__(self, frameCount=None, drxID=None, secondsCount=None, decimation=None, timeOffset=None, raw=None):
		self.frameCount = frameCount
		self.drxID = drxID
		self.secondsCount = secondsCount
		self.decimation = decimation
		self.timeOffset = timeOffset
		self.raw = raw
	
	def parseID(self):
		"""Parse the DRX ID into a tuple containing the beam (1 through
		4), tunning (1 and 2), and polarization (0 and 1)."""
		
		beam = self.drxID&7
		tune = (self.drxID>>3)&7
		pol  = (self.drxID>>7)&1

		return (beam, tune, pol)
	
	def getSampleRate(self):
		"""Return the sample rate of the data in samples/second."""
		
		sampleRate = dp_common.fS / self.decimation
		return sampleRate


class FrameData(object):
	"""Class that stores the information found in the data section of a DRX
	frame.  All three fields listed in the DP IDC version H are stored."""

	def __init__(self, timeTag=None, flags=None, iq=None):
		self.timeTag = timeTag
		self.flags = flags
		self.iq = iq

	def getTime(self):
		"""Function to convert the time tag from samples since station 
		midnight to seconds since station midnight.  This function needs 
		the dp_common module in order to work."""

		seconds = self.timeTag / dp_common.fS
		
		return seconds


class Frame(object):
	"""Class that stores the information contained within a single DRX 
	frame.  It's properties are FrameHeader and FrameData objects."""

	def __init__(self, header=FrameHeader(), data=FrameData()):
		self.header = header
		self.data = data
		self.valid = True

	def parseID(self):
		"""Convenience wrapper for the Frame.FrameHeader.parseID 
		function."""
		
		return self.header.parseID()

	def getSampleRate(self):
		"""Convenience wrapper for the Frame.FrameHeader.getSampleRate 
		function."""
		
		return self.header.getSampleRate()

	def getTime(self):
		"""Convenience wrapper for the Frame.FrameData.getTime function."""
		
		return self.data.getTime()
			
	def __add__(self, y):
		"""Add the data sections of two frames together or add a number 
		to every element in the data section."""
		
		try:
			return self.data.xy + y.self.data.xy
		except AttributeError:
			return self.data.xy + y
			
	def __iadd__(self, y):
		"""In-place add the data sections of two frames together or add 
		a number to every element in the data section."""
		
		try:
			self.data.xy += y.self.data.xy
		except AttributeError:
			self.data.xy += y
		
	def __mul__(self, y):
		"""Multiple the data sections of two frames together or multiply 
		a number to every element in the data section."""
		
		try:
			return self.data.xy * y.self.data.xy
		except AttributeError:
			return self.data.xy * y
			
	def __imul__(self, y):
		"""In-place multiple the data sections of two frames together or 
		multiply a number to every element in the data section."""
		
		try:
			self.data.xy *= y.self.data.xy
		except AttributeError:
			self.data.xy *= y
			
	def __eq__(self, y):
		"""Check if the time tags of two frames are equal or if the time
		tag is equal to a particular value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX == tY:
			return True
		else:
			return False
			
	def __ne__(self, y):
		"""Check if the time tags of two frames are not equal or if the time
		tag is not equal to a particular value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX != tY:
			return True
		else:
			return False
			
	def __gt__(self, y):
		"""Check if the time tag of the first frame is greater than that of a
		second frame or if the time tag is greater than a particular value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX > tY:
			return True
		else:
			return False
			
	def __ge__(self, y):
		"""Check if the time tag of the first frame is greater than or equal to 
		that of a second frame or if the time tag is greater than a particular 
		value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX >= tY:
			return True
		else:
			return False
			
	def __lt__(self, y):
		"""Check if the time tag of the first frame is less than that of a
		second frame or if the time tag is greater than a particular value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX < tY:
			return True
		else:
			return False
			
	def __le__(self, y):
		"""Check if the time tag of the first frame is less than or equal to 
		that of a second frame or if the time tag is greater than a particular 
		value."""
		
		tX = self.data.timeTag
		try:
			tY = self.data.timeTag
		except AttributeError:
			tY = y
		
		if tX <= tY:
			return True
		else:
			return False
			
	def __cmp__(self, y):
		"""Compare two frames based on the time tags.  This is helpful for 
		sorting things."""
		
		tX = self.data.timeTag
		tY = y.data.timeTag
		if tY > tX:
			return -1
		elif tX > tY:
			return 1
		else:
			return 0


class ObservingBlock(object):
	"""Class that stores all frames associates with a particular beam at a
	particular time."""

	def __init__(self, x1=Frame(), y1=Frame(), x2=Frame(), y2=Frame()):
		self.x1 = x1
		self.y1 = y1
		self.x2 = x2
		self.y2 = y2
		

def __readHeader(filehandle, Verbose=False):
	"""Private function to read in a DRX header.  Returns a FrameHeader object."""

	rawHeader = ''
	try:
		s = filehandle.read(4)
		rawHeader = rawHeader + s
		sync4, sync3, sync2, sync1 = struct.unpack(">BBBB", s)
		s = filehandle.read(4)
		rawHeader = rawHeader + s
		m5cID, frameCount3, frameCount2, frameCount1 = struct.unpack(">BBBB", s)
		frameCount = (long(frameCount3)<<16) | (long(frameCount2)<<8) | long(frameCount1)
		s = filehandle.read(4)
		rawHeader = rawHeader + s
		secondsCount = struct.unpack(">L", s)
		s = filehandle.read(4)
		rawHeader = rawHeader + s
		decimation, timeOffset = struct.unpack(">HH", s)
	except IOError:
		raise eofError()
	except struct.error:
		raise eofError()

	if sync1 != 92 or sync2 != 222 or sync3 != 192 or sync4 != 222:
		raise syncError(sync1=sync1, sync2=sync2, sync3=sync3, sync4=sync4)

	drxID = m5cID
	newHeader = FrameHeader()
	newHeader.frameCount = frameCount
	newHeader.drxID = drxID
	newHeader.secondsCount = secondsCount
	newHeader.decimation = decimation
	newHeader.timeOffset = timeOffset
	newHeader.raw = rawHeader

	if Verbose:
		beam, tune, pol = newHeader.parseID()
		print "Header: ", drxID, decimation, timeOffset
		print "  Beam: ", beam
		print "  Tuning: ", tune
		print "  Polarization: ", pol

	return newHeader


def __readData(filehandle):
	"""Private function to read in a DRX frame data section.  Returns a 
	FrameData object."""

	try:
		s = filehandle.read(8)
		timeTag = struct.unpack(">Q", s)
		s = filehandle.read(8)
		flags = struct.unpack(">Q", s)
	except IOError:
		raise eofError()
	except struct.error:
		raise eofError()
	
	# A truly excellent idea from Dan Wood
	rawData = numpy.fromfile(filehandle, dtype=numpy.uint8, count=4096)
	data = numpy.zeros(4096, dtype=numpy.complex_)
	if rawData.shape[0] < data.shape[0]:
		raise numpyError()
	
	data.real = (rawData>>4)&15
	data.imag = rawData&15

	# The data are signed, so apply the two-complement rule to generate 
	# the negative values
	negativeValues = numpy.where( data.real >= 8 )
	data.real[negativeValues] -= 16
	negativeValues = numpy.where( data.imag >= 8 )
	data.imag[negativeValues] -= 16
	
	newData = FrameData()
	newData.timeTag = timeTag[0]
	newData.flags = flags
	newData.iq = data

	return newData


def readFrame(filehandle, Verbose=False):
	"""Function to read in a single DRX frame (header+data) and store the 
	contents as a Frame object.  This function wraps readerHeader and 
	readData."""
	
	try:
		hdr = __readHeader(filehandle, Verbose=Verbose)
	except syncError, err:
		# Why?  If we run into a sync error here, then the following frame is invalid.  
		# Thus, we need to skip over this frame be advancing the file pointer 8+8+4096 B 
		currPos = filehandle.tell()
		frameEnd = currPos + FrameSize - 16
		filehandle.seek(frameEnd)
		raise err


	dat = __readData(filehandle)
	
	# Create the new frame object and return
	newFrame = Frame()
	newFrame.header = hdr
	newFrame.data = dat

	return newFrame


def readBlock(filehandle):
	"""Function to read in a single DRX block (four frames) and store the 
	contents as a ObservingBlock object.  This function wraps 
	readFrame."""
	
	try:
		x1 = readFrame(filehandle)
		y1 = readFrame(filehandle)
		x2 = readFrame(filehandle)
		y2 = readFrame(filehandle)
	except baseReaderError, err:
		raise err

	block = ObservingBlock(x1=x1, y1=y1, x2=x2, y2=y2)
	return block


def getBeamCount(filehandle):
	"""Find out how many beams are present by examining the first 16 DRX
	records.  Return the number of beams found."""

	# Save the current position in the file so we can return to that point
	fhStart = filehandle.tell()
	
	# Go back to the beginning...
	filehandle.seek(0)

	# Build up the list-of-lists that store ID codes and loop through 32
	# frames.  In each case, parse pull the DRX ID, extract the beam number, 
	# and append the DRX ID to the relevant beam array if it is not already 
	# there.
	beams = []
	for i in range(16):
		cFrame = readFrame(filehandle)
		cID = cFrame.header.drxID
		beam = cID&7
		if beam not in beams:
			beams.append(beam)
			
	# Return to the place in the file where we started
	filehandle.seek(fhStart)

	# Return the number of beams found
	return len(beams)


def getFramesPerObs(filehandle):
	"""Find out how many frames are present per beam by examining the first 
	16 DRX records.  Return the number of frames per observations as a four-
	element tuple, one for each beam."""
	
	# Save the current position in the file so we can return to that point
	fhStart = filehandle.tell()

	# Go back to the beginning...
	filehandle.seek(0)
	
	# Build up the list-of-lists that store ID codes and loop through 32
	# frames.  In each case, parse pull the DRX ID, extract the beam number, 
	# and append the DRX ID to the relevant beam array if it is not already 
	# there.
	idCodes = [[], [], [], []]
	for i in range(16):
		cFrame = readFrame(filehandle)
		cID = cFrame.header.drxID
		beam = cID&7
		if cID not in idCodes[beam-1]:
			idCodes[beam-1].append(cID)
			
	# Return to the place in the file where we started
	filehandle.seek(fhStart)
	
	# Get the length of each beam list and return them as a tuple
	return (len(idCodes[0]), len(idCodes[1]), len(idCodes[2]), len(idCodes[3]))
	

def averageObservations(Observations):
	"""Given a list of ObservingBlock objects, average the observations 
	together on a per tuning, per polarization basis.  A new ObsevingBlock 
	object is returned that contains the averaged data."""
	
	newBlock = Observations[0]
	tempx1 = newBlock.x1.data.iq * complex(0.0, 0.0)
	tempy1 = newBlock.y1.data.iq * complex(0.0, 0.0)
	tempx2 = newBlock.x2.data.iq * complex(0.0, 0.0)
	tempy2 = newBlock.y2.data.iq * complex(0.0, 0.0)
	
	obsCount = 0
	for Observation in Observations:
		tempx1 = tempx1 + Observation.x1.data.iq
		tempy1 = tempy1 + Observation.y1.data.iq
		tempx2 = tempx2 + Observation.x2.data.iq
		tempy2 = tempy2 + Observation.y2.data.iq
		obsCount = obsCount + 1
	
	tempx1 = tempx1 / float(obsCount)
	tempy1 = tempy1 / float(obsCount)
	tempx2 = tempx2 / float(obsCount)
	tempy2 = tempy2 / float(obsCount)
			
	newBlock.x1.data.iq = tempx1
	newBlock.y1.data.iq = tempy1
	newBlock.x2.data.iq = tempx2
	newBlock.y2.data.iq = tempy2
			
	return newBlock


def averageObservations2(Observations, timeAvg=1, chanAvg=1):
	"""Given a list of ObservingBlock objects, average the observations 
	together on a per tuning, per polarization basis.  A new ObsevingBlock 
	object is returned that contains the averaged data."""
	
	caBlocks = []
	if chanAvg > 1:
		for Observation in Observations:
			currBlock = Observation
			for i in range(len(Observation.x1.data.iq)/chanAvg):
				currBlock.x1.data.iq[i] = numpy.average( Observation.x1.data.iq[i*chanAvg:(i+1)*chanAvg] )
				currBlock.y1.data.iq[i] = numpy.average( Observation.y1.data.iq[i*chanAvg:(i+1)*chanAvg] )
				currBlock.x2.data.iq[i] = numpy.average( Observation.x2.data.iq[i*chanAvg:(i+1)*chanAvg] )
				currBlock.y2.data.iq[i] = numpy.average( Observation.y2.data.iq[i*chanAvg:(i+1)*chanAvg] )
			currBlock.x1.data.iq = currBlock.x1.data.iq[0:len(Observation.x1.data.iq)/chanAvg]
			currBlock.y1.data.iq = currBlock.y1.data.iq[0:len(Observation.x1.data.iq)/chanAvg]
			currBlock.x2.data.iq = currBlock.x2.data.iq[0:len(Observation.x1.data.iq)/chanAvg]
			currBlock.y2.data.iq = currBlock.y2.data.iq[0:len(Observation.x1.data.iq)/chanAvg]
			caBlocks.append(currBlock)
	else:
		caBlocks = Observations
			
	taBlocks = []
	if timeAvg > 1:
		for i in range(len(Observations)/timeAvg):
			taBlocks.append( averageObservations(caBlocks[i*timeAvg:(i+1)*timeAvg]) )
	else:
		taBlocks = caBlocks
			
	return taBlocks



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Read and plot the NOSTA_MAPPER table in a FITS IDI file writen by 
lsl.writer.fitsidi if it exists."""

import sys
import numpy
import pyfits

from lsl.correlator import uvUtils

import matplotlib.pyplot as plt
from matplotlib.ticker import NullFormatter


def main(args):
	# Open the FITS file for reading
	hdu = pyfits.open(args[0])

	# This try...except block catches files that don't use the stand mapper 
	# feature in lsl.writer.fitsidi.  If a NOSTA_MAPPER table doesn't exist,
	# the script fall back on the ARRAY_GEOMETRY table to gather the stand 
	# numbers.
	try:
		mapper = hdu['NOSTA_MAPPER']
		
		nosta = mapper.data.field('NOSTA')
		noact = mapper.data.field('NOACT')
		anname = mapper.data.field('ANNAME')
	except KeyError:
		ag = hdu['ARRAY_GEOMETRY']

		nosta = ag.data.field('NOSTA')
		noact = ag.data.field('NOSTA')
		anname = ag.data.field('ANNAME')

	# Print the stand mapping out
	print "Stand Mapping:"
	for sta,act,name in zip(nosta, noact, anname):
		print "  %3i -> %s (stand %3i)" % (sta, name, act)
	
	# Load in the positions of all of the stands
	xyz = uvUtils.getXYZ(numpy.arange(1,259))

	# Begin the plot
	fig = plt.figure()
	ax1 = fig.add_subplot(1, 1, 1)
	ax2 = plt.axes([0.75, 0.75, 0.1, 0.1])
	## Part 1 - Station between -80 and 80 (the section inside the fence)
	c = ax1.scatter(xyz[:,0], xyz[:,1], c=xyz[:,2], s=40.0, alpha=0.50)
	ax1.plot(xyz[noact-1,0], xyz[noact-1,1], marker='x', linestyle=' ', alpha=1.0, color='k')
	for m,a in zip(nosta, noact):
		ax1.text(xyz[a-1,0], xyz[a-1,1], '%i' % m)
	ax1.set_xlabel('$\Delta$X [E-W; m]')
	ax1.set_xlim([-80, 80])
	ax1.set_ylabel('$\Delta$Y [N-S; m]')
	ax1.set_ylim([-80, 80])
	## Part 2 - The outlier as inset axes
	ax2.scatter(xyz[:,0], xyz[:,1], c=xyz[:,2], s=40.0, alpha=0.50)
	ax2.plot(xyz[noact-1,0], xyz[noact-1,1], marker='x', linestyle=' ', alpha=1.0, color='k')
	for m,a in zip(nosta, noact):
		ax2.text(xyz[a-1,0], xyz[a-1,1], '%i' % m)
	ax2.set_title('Outlier')
	ax2.set_xlim([335, 345])
	ax2.set_ylim([10, 20])
	ax2.xaxis.set_major_formatter( NullFormatter() )
	ax2.yaxis.set_major_formatter( NullFormatter() )

	plt.show()

if __name__ == "__main__":
	main(sys.argv[1:])



# -*- coding: utf-8 -*-

"""Python module for creating creating, validating, and writing simulated 
TBN frames to a file."""

import numpy

from lsl.common import dp as dp_common
from lsl.reader import tbn
from errors import *

__version__ = '0.1'
__revision__ = '$ Revision: 6 $'
__all__ = ['SimFrame', 'frame2frame', '__version__', '__revision__', '__all__']


def frame2frame(tbnFrame):
	"""Convert a tbn.Frame object to a raw DP TBN frame."""

	# The raw frame
	rawFrame = numpy.zeros(tbn.FrameSize, dtype=numpy.uint8)

	# Part 1: The header
	## Sync. words (0xDEC0DE5C)
	rawFrame[0] = 0xDE  # 222
	rawFrame[1] = 0xC0  # 192
	rawFrame[2] = 0xDE  # 222
	rawFrame[3] = 0x5C  #  92
	## Frame count
	rawFrame[5] = (tbnFrame.header.frameCount>>16) & 255
	rawFrame[6] = (tbnFrame.header.frameCount>>8) & 255
	rawFrame[7] = tbnFrame.header.frameCount & 255
	## Seconds count
	rawFrame[8] = (tbnFrame.header.secondsCount>>24) & 255
	rawFrame[9] = (tbnFrame.header.secondsCount>>16) & 255
	rawFrame[10] = (tbnFrame.header.secondsCount>>8) & 255
	rawFrame[11] = tbnFrame.header.secondsCount & 255
	## TBN ID
	rawFrame[12] = (tbnFrame.header.tbnID>>8) & 255
	rawFrame[13] = tbnFrame.header.tbnID & 255
	## NB: Next two bytes are unsigned
	
	# Part 2: The data
	## Time tag
	rawFrame[16] = (tbnFrame.data.timeTag>>56) & 255
	rawFrame[17] = (tbnFrame.data.timeTag>>48) & 255
	rawFrame[18] = (tbnFrame.data.timeTag>>40) & 255
	rawFrame[19] = (tbnFrame.data.timeTag>>32) & 255
	rawFrame[20] = (tbnFrame.data.timeTag>>24) & 255
	rawFrame[21] = (tbnFrame.data.timeTag>>16) & 255
	rawFrame[22] = (tbnFrame.data.timeTag>>8) & 255
	rawFrame[23] = tbnFrame.data.timeTag & 255
	## Data
	i = tbnFrame.data.iq.real
	q = tbnFrame.data.iq.imag
	### Round, clip, and convert to unsigned integers
	i = i.round()
	i = i.clip(-128, 127)
	i = i.astype(numpy.uint8)
	q = q.round()
	q = q.clip(-128, 127)
	q = q.astype(numpy.uint8)
	
	rawFrame[24::2] = i
	rawFrame[25::2] = q
	
	return rawFrame


class SimFrame(tbn.Frame):
	"""tbn.SimFrame extends the lsl.reader.tbn.Frame object to yield a method 
	for easily creating DP ICD-compliant raw TBN frames.  Frames created with
	this method can be written to a file via the methods writeRawFrame() function."""

	def __init__(self, stand=None, pol=None, frameCount=None, obsTime=None, iq=None):
		"""Given a list of parameters, build a tbn.SimFrame object.  The parameters
		needed are:
		  + stand id (>0 & <259)
		  + polarization (0 for x, or 1 for y)
		  + which frame number to create
		  + observation time in seconds since the epoch
		  + 1-D numpy array representing the frame I/Q (complex) data
		Not all of these parameters are needed at initialization of the object and
		the values can be added later."""
		
		self.stand = stand
		self.pol = pol
		self.frameCount = frameCount
		self.obsTime = obsTime
		self.iq = iq
		
		self.header = tbn.FrameHeader()
		self.data = tbn.FrameData()
		
	def __update(self):
		"""Private function to use the object's parameter values to build up 
		a tbn.Frame-like object."""
		
		self.header.frameCount = self.frameCount
		self.header.secondsCount = int(self.obsTime)
		self.header.tbnID = 2*(self.stand-1) + self.pol + 1
		
		self.data.timeTag = long(self.obsTime * dp_common.fS)
		self.data.iq = self.iq
	
	def loadFrame(self, tbnFrame):
		"""Populate the a tbn.SimFrame object with a pre-made frame."""
		
		self.header = tbnFrame.header
		self.data = tbnFrame.data
		
		# Back-fill the class' fields to make sure the object is consistent
		## Header
		self.stand = self.header.parseID()[0]
		self.pol = self.header.parseID()[1]
		self.frameCount = self.header.frameCount
		## Data
		self.obsTime = self.data.timeTag / dp_common.fS
		self.iq = self.data.iq
	
	def isValid(self, raiseErrors=False):
		"""Check if simulated TBN frame is valid or not.  Valid frames return 
		True and invalid frames False.  If the 'raiseErrors' keyword is set, 
		isValid raises an error when a problem is encountered."""

		# Make sure we have the latest values
		self.__update()

		stand, pol = self.parseID()
		# Is the stand number reasonable?
		if stand == 0 or stand > 258:
			if raiseErrors:
				raise invalidStand()
			return False

		# Is the polarization reasonable?
		if pol not in [0, 1]:
			if raiseErrors:
				raise invalidPol()
			return False

		# Is there data loaded into frame.data.iq?
		if self.data.iq is None:
			if raiseErrors:
				raise invalidDataSize()
			return False

		# Does the data length make sense?
		if self.data.iq.shape[0] != 512:
			if raiseErrors:
				raise invalidDataSize()
			return False

		# Does the data type make sense?
		if self.data.iq.dtype.kind != 'c':
			if raiseErrors:
				raise invalidDataType()
			return False

		# If we made it this far, it's valid
		return True

	def createRawFrame(self):
		"""Re-express a simulated TBN frame as a numpy array of unsigned 8-bit 
		integers.  Returns a numpy array if the frame  is valid.  If the frame 
		is not ICD-compliant, a errors.baseSimError-type error is raised."""

		# Make sure we have the latest values
		self.__update()

		self.isValid(raiseErrors=True)
		return frame2frame(self)

	def writeRawFrame(self, fh):
		"""Write a simulated TBN frame to a filehandle if the frame is valid.
		If the frame is not ICD-compliant, a errors.baseSimError-type error 
		is raised."""

		# Make sure we have the latest values
		self.__update()

		rawFrame = self.createRawFrame()
		rawFrame.tofile(fh)

	def __str__(self):
		if self.stand is None:
			return "Empty TBN SimFrame object"
		else:
			return "TBN SimFrame for stand %i, pol. %i @ time %i" % (self.stand, self.pol, self.obsTime)



# -*- coding: utf-8 -*-

"""Unit test suite for the lsl package."""

import unittest

from lsl.tests import test_paths
from lsl.tests import test_astro
from lsl.tests import test_skymap
from lsl.tests import test_mathutil
from lsl.tests import test_nec_util
from lsl.tests import test_catalog
from lsl.tests import test_stations
from lsl.tests import test_robust
from lsl.tests import test_reader
from lsl.tests import test_uvUtils
from lsl.tests import test_fx
from lsl.tests import test_visUtils
from lsl.tests import test_fakedata
from lsl.tests import test_simdp
from lsl.tests import test_simvis
from lsl.tests import test_geodesy
from lsl.tests import test_fitsidi
from lsl.tests import test_tsfits
from lsl.tests import test_sdfits
from lsl.tests import test_vdif
from lsl.tests import test_beamformer


__revision__  = "$Revision: 95 $"
__version__   = "0.1"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


class lsl_tests(unittest.TestSuite):
	"""A unittest.TestSuite class which contains all of the package unit tests."""

	def __init__(self):
		"""Setup the lsl package unit test suite."""

		unittest.TestSuite.__init__(self)
		
		self.addTest(test_paths.paths_test_suite())
		self.addTest(test_astro.astro_test_suite())
		self.addTest(test_skymap.skymap_test_suite())
		self.addTest(test_mathutil.mathutil_test_suite())
		self.addTest(test_nec_util.nec_util_test_suite())
		self.addTest(test_catalog.catalog_test_suite())
		self.addTest(test_stations.stations_test_suite())
		self.addTest(test_robust.robust_test_suite())
		self.addTest(test_reader.reader_test_suite())
		self.addTest(test_uvUtils.uvUtils_test_suite())
		self.addTest(test_fx.fx_test_suite())
		self.addTest(test_visUtils.visUtils_test_suite())
		self.addTest(test_fakedata.fakedata_test_suite())
		self.addTest(test_simdp.simdp_test_suite())
		self.addTest(test_simvis.simvis_test_suite())
		self.addTest(test_geodesy.geodesy_test_suite())
		self.addTest(test_fitsidi.fitsidi_test_suite())
		self.addTest(test_tsfits.tsfits_test_suite())
		self.addTest(test_sdfits.sdfits_test_suite())
		self.addTest(test_vdif.vdif_test_suite())
		self.addTest(test_beamformer.beamformer_test_suite())


def main(opts=None, args=None):
	"""Function to call all of the lsl tests."""

	if opts is not None:
		if opts.verbose:
			level = 2
		else:
			level = 1
	else:
		level = 2
			
	suite = lsl_tests()
	runner = unittest.TextTestRunner(verbosity = level)
	runner.run(suite)


if __name__  == '__main__':
	import optparse
	
	parser = optparse.OptionParser(usage = "python %prog [options]", description = __doc__)
	parser.add_option("-v", "--verbose", action = "store_true", dest = "verbose", default = False,
		help = "extra print output")
	(opts, args) = parser.parse_args()

	main(opts, args)



# -*- coding: utf-8 -*-

"""Modules to take TBW/TBN time series data and write it to a FITS file composed 
of binary tables.  The format of TSFITS files is:

PRIMARY HDU
  Keywords:
    * OBJECT - object being observed in this data set, e.g., 'zenith'
    * TELESCOP - telescope used for the observations, e.g., 'LWA-1'
    * OBSMODE - observation mode used for the data.  Options are:
        1. TBW
        2. TBN
    * NSTANDS - number of stands found in the file
    * TBWBITS *optional* - how many bits are used by the TBW data (4 or 12)
    * FILTER *optional* - TBN filter code used to acquire the data
    * GAIN *optional* - TBN gain setting used to acquire the data
    * FREQ *optional* - central frequency in Hz for the TBN observations

TIME SERIES HDU
  Binary table that stores the actual observations.  There is one TIME SERIES 
  extionsion per stand in the data.
  Keywords:
    * EXTNAME - extension name of 'TIME SERIES'
    * EXTVER - extension version equal to the order in which the data was 
      added to the FITS file
    * STAND - stand number associated with this data
    * DATE-OBS - date of observation for the first sample of the first data 
      row

  Data:
    * DATA - column storing the observations.  For TBW data this consists of 
      400 (12-bit) or 1200 (4-bit) elements per row stored as 16-bit integers.
      For TBN data this consits of 512 elements per row stored as 32-bit 
      complex numbers.
    * POL - column storing the polarization associated with the DATA column.  
      'x' polarization is stored as 0 and 'y' polarization is stored as 1.  
      The values is stored as a 16-bit integer.
    * TIME - number of samples at f_S (196 MSamples/s) since DATE-OBS for the
      first sample in the row.  This is stored as a 64-bit integer.
"""

import os
import sys
import time
import numpy
import pyfits
from datetime import datetime, timedelta, tzinfo

from lsl.common import dp as dp_common
from lsl.reader.tbn import filterCodes as tbnCodes
from lsl.common.warns import warnDeprecated

__version__ = '0.5'
__revision__ = '$ Revision: 16 $'
__all__ = ['UTC', 'TSFITS', 'TBW', 'TBN', '__version__', '__revision__', '__all__']


class UTC(tzinfo):
    """tzinfo object for UTC time."""

    def utcoffset(self, dt):
        return timedelta(0)

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return timedelta(0)


class TSFITS(object):
	"""Class that holds TSFITS data until it is ready to be writen to disk."""

	def __init__(self, filename, mode, Overwrite=False, UseQueue=True, verbose=False):
		"""Initialize a TSFITS object using a filename and an observation mode 
		(TBW or TBN).  Optionally, TSFITS can be told to overwrite the file if it 
		already exists using the 'Overwrite' keyword."""

		assert(mode in ['TBW', 'TBN'])

		self.filename = filename
		self.mode = mode
		self.hdulist = None
		self.UseQueue = UseQueue
		self.queue = {}
		self.queueLimit = 10000
		self.site = 'LWA'
		self.firstSamples = {}
		self.verbose = verbose

		if os.path.exists(filename) and not Overwrite:
			self.hdulist = pyfits.open(self.filename, mode="update", memmap=0)
			self.standCount = self.hdulist[0].header['nstand']
		else:
			if os.path.exists(filename):
				os.path.delete(filename)

			self.standCount = 0
			primary = pyfits.PrimaryHDU()
			primary.header.update('OBJECT', 'zenith')
			primary.header.update('TELESCOP', self.site)
			primary.header.update('OBSMODE', self.mode)
			primary.header.update('NSTAND', self.standCount)

			hdulist = pyfits.HDUList([primary])
			hdulist.writeto(filename)

			self.hdulist = pyfits.open(self.filename, mode="update", memmap=0)

	def info(self):
		"""Short-cut to the pyfits.info() function on an opened FITS file."""

		self.hdulist.info()

	def flush(self):
		"""Short-cut to the pyfits.flush() function on an opened FITS file."""

		self.hdulist.flush()

	def close(self):
		"""Empty the data queue (if it exists) and write all changes to disk 
		using flush."""

		if self.UseQueue:
			self.__emptyQueue()
		self.flush()

	def setSite(self, site):
		"""Set the TELESCOP keyword in the primary HDU using an lsl.common.stations
		object."""

		self.site = site.name

		self.hdulist[0].header.update('TELESCOP', self.site)

		self.flush()

	def __findExtension(self, stand):
		"""Private function to find out which extension stores the stand in 
		question.  None is returned is that stand is not in the TSFITS file."""

		extension = None

		i = 1
		for hdu in self.hdulist[1:]:
			if int(hdu.header['stand']) == int(stand):
				extension = i
				break
			i = i + 1

		return extension

	def __makeAppendTable(self, extension, AddRows=1):
		"""Private function to make a temporary table for appending data."""

		nrows = self.hdulist[extension].data.shape[0]
		tempHDU = pyfits.new_table(self.hdulist[extension].columns, nrows=nrows+AddRows)
		tempHDU.header = self.hdulist[extension].header.copy()
	
		return tempHDU

	def __applyAppendTable(self, extension, tempHDU):
		"""Private function to replace the given extension with the temporary
		table."""

		self.hdulist[extension] = tempHDU
		self.flush()

	def __addDataSingle(self, frame):
		"""Private function to add a single data entry to a TSFITS file.  This 
		method is not particular fast since the existing file needs to be copied
		to append data to the end of a particular extension."""

		if self.mode == 'TBW':
			stand = frame.parseID()
		else:
			stand, pol = frame.parseID()
		extension = self.__findExtension(stand)

		if extension is None:
			if self.verbose:
				print "Stand '%i' not found, creating new binary table extension" % stand
			self.standCount = self.standCount + 1

			if self.mode == 'TBW':
				# Data
				c1 = pyfits.Column(name='data', format='%iI' % frame.data.xy.shape[1], array=frame.data.xy.astype(numpy.int16))
				# Polarization
				c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0, 1], dtype=numpy.int16))
				# Time
				c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0, 0]))
			else:
				# Data
				data = frame.data.iq.astype(numpy.csingle)
				data.shape = (1,512)
				c1 = pyfits.Column(name='data', format='512C', array=data)
				# Polarization
				c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0], dtype=numpy.int16))
				# Time
				c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0]))

			# Define the collection of columns
			colDefs = pyfits.ColDefs([c1, c2, c3])

			# Get the time of the first sample and convert it to a datetime
			self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
			firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

			tsfits = pyfits.new_table(colDefs)
			tsfits.header.update('EXTNAME', 'TIME SERIES', after='tfields')
			tsfits.header.update('EXTVER', self.standCount, after='EXTNAME')
			tsfits.header.update('STAND', stand, after='EXTVER')
			tsfits.header.update('DATE-OBS', firstSample.isoformat())

			self.hdulist.append(tsfits)
			self.hdulist[0].header.update('NSTAND', self.standCount)
			self.flush()
			
			self.hdulist[-1].data.field('pol')[0] = numpy.array([0], dtype=numpy.int16)
			self.hdulist[-1].data.field('time')[0] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
			if self.mode == 'TBW':
				self.hdulist[-1].data.field('pol')[1] = numpy.array([1], dtype=numpy.int16)
				self.hdulist[-1].data.field('time')[1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
		else:
			# Make sure that we have a first sample time to reference to if 
			# we are adding on to the end of a file
			if stand not in self.firstSamples.keys():
				firstSample = datetime.strptime(self.hdulist[extension].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S")
				self.firstSamples[stand] = long(time.mktime(firstSample.timetuple()))

			nrows = self.hdulist[extension].data.shape[0]
			if self.mode == 'TBW':
				tempHDU = self.__makeAppendTable(extension, AddRows=2)
				tempHDU.data.field('data')[nrows] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[0,:])
				tempHDU.data.field('pol')[nrows] = numpy.array([0], dtype=numpy.int16)
				tempHDU.data.field('time')[nrows] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
				tempHDU.data.field('data')[nrows+1] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[1,:])
				tempHDU.data.field('pol')[nrows+1] = numpy.array([1], dtype=numpy.int16)
				tempHDU.data.field('time')[nrows+1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
			else:
				tempHDU = self.__makeAppendTable(extension, AddRows=1)
				data = frame.data.iq.astype(numpy.csingle)
				data.shape = (1,512)
				tempHDU.data.field('data')[nrows:] = data
				tempHDU.data.field('pol')[nrows:] = numpy.array([pol], dtype=numpy.int16)
				tempHDU.data.field('time')[nrows:] = numpy.array([frame.data.timeTag]) / dp_common.fS - self.firstSamples[stand]

			self.__applyAppendTable(extension, tempHDU)

		self.flush()

	def __addDataQueue(self, frame):
		"""Private function similar to __addDataSignle, but it saves the data to 
		memory (self.queue) to be written once the queue fills up."""

		if self.mode == 'TBW':
			stand = frame.parseID()
		else:
			stand, pol = frame.parseID()
		if stand not in self.queue.keys():
			self.queue[stand] = []
		self.queue[stand].append(frame)

		if len(self.queue[stand]) >= self.queueLimit:
			start = 0
			extension = self.__findExtension(stand)

			if extension is None:
				start = 1
				frame = self.queue[stand][0]
				
				if self.verbose:
					print "Stand '%i' not found, creating new binary table extension" % stand
				self.standCount = self.standCount + 1

				if self.mode == 'TBW':
					# Data
					c1 = pyfits.Column(name='data', format='%iI' % frame.data.xy.shape[1], array=frame.data.xy.astype(numpy.int16))
					# Polarization
					c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0, 1], dtype=numpy.int16))
					# Time
					c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0, 0]))
				else:
					# Data
					data = frame.data.iq.astype(numpy.csingle)
					data.shape = (1,512)
					c1 = pyfits.Column(name='data', format='512C', array=data)
					# Polarization
					c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0], dtype=numpy.int16))
					# Time
					c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0]))

				# Define the collection of columns
				colDefs = pyfits.ColDefs([c1, c2, c3])

				# Get the time of the first sample and convert it to a datetime
				self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
				firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

				tsfits = pyfits.new_table(colDefs)
				tsfits.header.update('EXTNAME', 'TIME SERIES', after='tfields')
				tsfits.header.update('EXTVER', self.standCount, after='EXTNAME')
				tsfits.header.update('STAND', stand, after='EXTVER')
				tsfits.header.update('DATE-OBS', firstSample.isoformat('T'))

				self.hdulist.append(tsfits)
				self.hdulist[0].header.update('NSTAND', self.standCount)
				self.flush()
				
				self.hdulist[-1].data.field('pol')[0] = numpy.array([0], dtype=numpy.int16)
				self.hdulist[-1].data.field('time')[0] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
				if self.mode == 'TBW':
					self.hdulist[-1].data.field('time')[1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
					self.hdulist[-1].data.field('pol')[1] = numpy.array([1], dtype=numpy.int16)

				self.flush()
				extension = self.__findExtension(stand)

			# Make sure that we have a first sample time to reference to if 
			# we are adding on to the end of a file
			if stand not in self.firstSamples.keys():
				firstSample = datetime.strptime(self.hdulist[extension].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S")
				self.firstSamples[stand] = long(time.mktime(firstSample.timetuple()))

			nrows = self.hdulist[extension].data.shape[0]
			if self.mode == 'TBW':
				tempHDU = self.__makeAppendTable(extension, AddRows=2*(self.queueLimit-start))
				for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
					tempHDU.data.field('data')[nrows+2*count] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[0,:])
					tempHDU.data.field('pol')[nrows+2*count] = numpy.array([0], dtype=numpy.int16)
					tempHDU.data.field('time')[nrows+2*count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
					tempHDU.data.field('data')[nrows+2*count+1] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[1,:])
					tempHDU.data.field('pol')[nrows+2*count+1] = numpy.array([1], dtype=numpy.int16)
					tempHDU.data.field('time')[nrows+2*count+1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
			else:
				tempHDU = self.__makeAppendTable(extension, AddRows=(self.queueLimit-start))
				for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
					stand,pol = frame.parseID()
					data = frame.data.iq.astype(numpy.csingle)
					data.shape = (1,512)
					tempHDU.data.field('data')[nrows+count] = data
					tempHDU.data.field('pol')[nrows+count] = numpy.array([pol], dtype=numpy.int16)
					tempHDU.data.field('time')[nrows+count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]

			self.__applyAppendTable(extension, tempHDU)
			del(self.queue[stand])

	def __emptyQueue(self):
		"""Private function to empty a empty the self.queue dictionary on demand
		if it exists."""

		for stand in self.queue.keys():
			if len(self.queue[stand]) == 0:
				continue
			queueSize = len(self.queue[stand])
			start = 0
			extension = self.__findExtension(stand)

			if extension is None:
				start = 1
				frame = self.queue[stand][0]
				
				if self.verbose:
					print "Stand '%i' not found, creating new binary table extension" % stand
				self.standCount = self.standCount + 1

				if self.mode == 'TBW':
					# Data
					c1 = pyfits.Column(name='data', format='%iI' % frame.data.xy.shape[1], array=frame.data.xy.astype(numpy.int16))
					# Polarization
					c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0, 1], dtype=numpy.int16))
					# Time
					c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0, 0]))
				else:
					# Data
					data = frame.data.iq.astype(numpy.csingle)
					data.shape = (1,512)
					c1 = pyfits.Column(name='data', format='512C', array=data)
					# Polarization
					c2 = pyfits.Column(name='pol', format='1I', array=numpy.array([0], dtype=numpy.int16))
					# Time
					c3 = pyfits.Column(name='time', format='1K', array=numpy.array([0]))

				# Define the collection of columns
				colDefs = pyfits.ColDefs([c1, c2, c3])

				# Get the time of the first sample and convert it to a datetime
				self.firstSamples[stand] = long(frame.data.timeTag / dp_common.fS)
				firstSample = datetime.utcfromtimestamp(self.firstSamples[stand])

				tsfits = pyfits.new_table(colDefs)
				tsfits.header.update('EXTNAME', 'TIME SERIES', after='tfields')
				tsfits.header.update('EXTVER', self.standCount, after='EXTNAME')
				tsfits.header.update('STAND', stand, after='EXTVER')
				tsfits.header.update('DATE-OBS', firstSample.isoformat('T'))

				self.hdulist.append(tsfits)
				self.hdulist[0].header.update('NSTAND', self.standCount)
				self.flush()
				
				self.hdulist[-1].data.field('pol')[0] = numpy.array([0], dtype=numpy.int16)
				self.hdulist[-1].data.field('time')[0] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
				if self.mode == 'TBW':
					self.hdulist[-1].data.field('time')[1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
					self.hdulist[-1].data.field('pol')[1] = numpy.array([1], dtype=numpy.int16)

				self.flush()
				extension = self.__findExtension(stand)

			# Make sure that we have a first sample time to reference to if 
			# we are adding on to the end of a file
			if stand not in self.firstSamples.keys():
				firstSample = datetime.strptime(self.hdulist[extension].header['DATE-OBS'], "%Y-%m-%dT%H:%M:%S")
				self.firstSamples[stand] = long(time.mktime(firstSample.timetuple()))

			nrows = self.hdulist[extension].data.shape[0]
			if self.mode == 'TBW':
				tempHDU = self.__makeAppendTable(extension, AddRows=2*(queueSize-start))
				for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
					tempHDU.data.field('data')[nrows+2*count] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[0,:])
					tempHDU.data.field('pol')[nrows+2*count] = numpy.array([0], dtype=numpy.int16)
					tempHDU.data.field('time')[nrows+2*count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
					tempHDU.data.field('data')[nrows+2*count+1] = numpy.squeeze(frame.data.xy.astype(numpy.int16)[1,:])
					tempHDU.data.field('pol')[nrows+2*count+1] = numpy.array([1], dtype=numpy.int16)
					tempHDU.data.field('time')[nrows+2*count+1] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]
			else:
				tempHDU = self.__makeAppendTable(extension, AddRows=(queueSize-start))
				for count,frame in zip(range(len(self.queue[stand][start:])), self.queue[stand][start:]):
					stand,pol = frame.parseID()
					data = frame.data.iq.astype(numpy.csingle)
					data.shape = (1,512)
					tempHDU.data.field('data')[nrows+count] = data
					tempHDU.data.field('pol')[nrows+count] = numpy.array([pol], dtype=numpy.int16)
					tempHDU.data.field('time')[nrows+count] = frame.data.timeTag / dp_common.fS - self.firstSamples[stand]

			self.__applyAppendTable(extension, tempHDU)
			del(self.queue[stand])

	def addStandData(self, frame):
		"""Add a frame object to the TSFITS file.  This function takes care of 
		figuring out which extension the data goes to and how it should be formated.
		This function also updates the primary HDU with information about the data, 
		i.e., TBW data bits, TBN filter code, etc."""

		if self.mode == 'TBW':
			try:
				self.hdulist[0].header['TBWBITS']
			except:
				self.hdulist[0].header.update('TBWBITS', frame.getDataBits())
				self.flush()
		else:
			try:
				self.hdulist[0].header['FILTER']
			except:
				if frame.getFilterCode() is not None:
					self.hdulist[0].header.update('FILTER', frame.getFilterCode())
				if frame.data.gain is not None:
					self.hdulist[0].header.update('GAIN', frame.data.gain)
				if frame.data.centralFreq is not None:
					self.hdulist[0].header.update('FREQ', frame.data.centralFreq)
				self.flush()

		if self.UseQueue:
			self.__addDataQueue(frame)
		else:
			self.__addDataSingle(frame)

	def getStandData(self, stand):
		"""Retrieve a dictionary of all data stored for a particular stand.  The 
		dictionary keys are:
		  * *data* - numpy array of data
		  * *pol* - numpy array of polarizations
		  * *time* - numpy array of times in samples at f_S since DATE-OBS
		"""

		extension = self.__findExtension(stand)
		
		if extension is None:
			output = {}
			for key in ['data', 'pol', 'time']:
				output[key] = None
		else:
			output = {}
			for key in ['data', 'pol', 'time']:
				output[key] = self.hdulist[extension].data.field(key)
		return output


class TBW(TSFITS):
	"""Sub-class of TSFITS for dealing with TBW data in particular."""

	def __init__(self, filename, Overwrite=False, UseQueue=True, verbose=False):
		super(TBW, self).__init__(filename, 'TBW', Overwrite=Overwrite, UseQueue=UseQueue, verbose=verbose)
		

class TBN(TSFITS):
	"""Sub-class of TSFITS for dealing with TBN data in particular."""

	def __init__(self, filename, Overwrite=False, UseQueue=True, verbose=False):
		super(TBN, self).__init__(filename, 'TBN', Overwrite=Overwrite, UseQueue=UseQueue, verbose=verbose)



# -*- coding: utf-8 -*

"""lsl.reader - Modular readers for the various LWA data formats:
 * tbw
 * tbn
 * drx
 * s60

A ring buffer for re-ordering TBN/DRX data is included in the 'buffer'
module."""

import errors

import s60
import tbw
import tbn
import drx

import buffer



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""Predict driftcurve for a given site using a given antenna model."""

import sys
import optparse
import logging
import math
import os

import numpy
import pylab

from lsl import skymap, astro
from lsl.common import stations
from lsl.sim import nec_util
        

__revision__ = "$Revision: 93 $"
__version__  = "0.1"
__author__    = "D.L.Wood"
__maintainer__ = "Jayce Dowell"


if __name__ == '__main__':
	# setup logger

	logging.basicConfig()
	log = logging.getLogger()
	log.setLevel(logging.INFO)
	
	# parse command line
	parser = optparse.OptionParser(usage="%prog [options] FREQ NECFILENAME SKYMAPFILENAME", \
		description="Generate a drift curve for a dipole at SITE observing at a given FREQ (MHz).  SITE must be one of the sites known by the station module in lwda_util.")
	parser.add_option("-v", "--verbose", action = "store_true", dest = "verbose",
		default = False, help = "enable debug messages")
	parser.add_option("-x", "--doplot", action = "store_true", dest = "doplot",
		default = False, help = "Make an X-windows plot")
	parser.add_option("-s", "--site", action = "store", dest = "site", type = "str",
		default = "LWDA", help = "site name (default LWDA)")
	parser.add_option("-p", "--polarization", action = "store", dest = "polarization",
		default = "NS", help = "antenna polarization orientation (NS or EW)")
	
	(opts, args) = parser.parse_args()    

	if len(args) != 3:
		parser.error("wrong number of arguments")
		
	if opts.verbose:
		log.setLevel(logging.DEBUG)
		
	if (opts.polarization != 'EW') and (opts.polarization != 'NS'):
		parser.error("only 'EW' and 'NS' allowed for polarization")
		
	# Extract the command line arguments
	freq = float(args[0])
	if freq < 1.0 or freq > 1000.0:
		log.error("Please specify a frequency in MHz!")
		sys.exit(1)
	necname = args[1]
	skymapname = args[2]
	
	# get site geo position
	try:
		sta = stations.lwa1()
	except:
		log.error("cannot get info for site %s", site)
		sys.exit(1)
	log.info("Site %s, Long %f", sta.name, sta.long*180.0/math.pi)
	
	# read skymap and scale to target frequency
	log.info("Reading skymap file")
	smap = skymap.SkyMap(skymapname)
	log.info("Read skymap of %d x %d pixels, min=%f, max=%f",smap.numPixelsX,smap.numPixelsY,smap._power.min(),smap._power.max())
	
	log.info("Reading antenna info")    
	# get user-supplied pattern
	antPatAnt = nec_util.NECPattern(necname, freq).antenna_pat_dB
	#pylab.figure(3)
	log.info("Read pattern.  Min %f Max %f", antPatAnt.min(), antPatAnt.max())
	#pylab.imshow(antPatAnt.transpose(), origin='lower', vmin=-10.0, cmap=pylab.cm.hot)
	antPatAnt = numpy.power(10.0, antPatAnt / 10.0)
	
	# calculate times in both site LST and UTC
	t0 = astro.get_julian_from_sys()
	lst = astro.get_local_sidereal_time(sta.long*180.0/math.pi, t0) / 24.0
	t0 -= lst*(23.933/24.0) # Compensate for shorter sidereal days
	times = numpy.arange(0, 1, 0.2/24) + t0
	
	lstList = []
	powListAnt = [] 
	
	for t in times:
		# project skymap to site location and observation time
		pmap = skymap.ProjectedSkyMap(smap, sta.lat*180.0/math.pi, sta.long*180.0/math.pi, t)
		lst = astro.get_local_sidereal_time(sta.long*180.0/math.pi, t)
		iaz = pmap.visibleAz.astype(numpy.int_)
		ialt = pmap.visibleAlt.astype(numpy.int_)
		lstList.append(lst)
		
		#log.info("iaz : %f - %f",iaz.min(),iaz.max())
		#log.info("ialt : %f - %f",ialt.min(),ialt.max())
		if opts.polarization == 'EW':
			iaz += 90.0
			iaz = (iaz >= 360).choose(iaz, iaz - 360)
		
		cdec = numpy.cos(pmap.visibleDec * smap.degToRad)
				
		# convolution of user antenna pattern with visible skymap
		gain = antPatAnt[iaz, ialt]
		powerAnt = (pmap.visiblePower * gain * cdec).sum() / (gain * cdec).sum()
		powListAnt.append(powerAnt)
		
		log.debug("LST=%f, power_ant=%f", lst, powerAnt) 
			
	# plot results
	if opts.doplot:
		pylab.figure(1)
		pylab.title("Driftcurve: Site %s Pol %s Ant %s Freq %0.2f MHz" % \
			(opts.site, opts.polarization, os.path.basename(necname), freq))
		pylab.plot(lstList, powListAnt, "ro",label="NEC Antenna Pattern")
		pylab.xlabel("LST")
		pylab.ylabel("Temp (K)")
		pylab.grid(1)
		pylab.show()
	
	mf = file("model_%s_%s_%s_%0.2f.txt" % (opts.site, os.path.basename(necname),opts.polarization, freq),"w")
	for lst,pow in zip(lstList,powListAnt):
		mf.write("%f  %f\n" % (lst,pow))
	mf.close()
	
	sys.exit(0)



# -*- coding: utf-8 -*-

"""Python module for creating creating, validating, and writing simulated 
TBW frames to a file."""

import numpy

from lsl.common import dp as dp_common
from lsl.reader import tbw
from errors import *

__version__ = '0.2'
__revision__ = '$ Revision: 9 $'
__all__ = ['SimFrame', 'frame2frame', '__version__', '__revision__', '__all__']


def frame2frame(tbwFrame):
	"""Convert a tbw.Frame/tbw.SimFrame object to a raw DP TBW frame."""

	# The raw frame
	rawFrame = numpy.zeros(tbw.FrameSize, dtype=numpy.uint8)

	# Part 1: The header
	## Sync. words (0xDEC0DE5C)
	rawFrame[0] = 0xDE  # 222
	rawFrame[1] = 0xC0  # 192
	rawFrame[2] = 0xDE  # 222
	rawFrame[3] = 0x5C  #  92
	## Frame count
	rawFrame[5] = (tbwFrame.header.frameCount>>16) & 255
	rawFrame[6] = (tbwFrame.header.frameCount>>8) & 255
	rawFrame[7] = tbwFrame.header.frameCount & 255
	## Seconds count
	rawFrame[8] = (tbwFrame.header.secondsCount>>24) & 255
	rawFrame[9] = (tbwFrame.header.secondsCount>>16) & 255
	rawFrame[10] = (tbwFrame.header.secondsCount>>8) & 255
	rawFrame[11] = tbwFrame.header.secondsCount & 255
	## TBW ID
	rawFrame[12] = (tbwFrame.header.tbwID>>8) & 255
	rawFrame[13] = tbwFrame.header.tbwID & 255
	## NB: Next two bytes are unsigned
	
	# Part 2: The data
	## Time tag
	rawFrame[16] = (tbwFrame.data.timeTag>>56) & 255
	rawFrame[17] = (tbwFrame.data.timeTag>>48) & 255
	rawFrame[18] = (tbwFrame.data.timeTag>>40) & 255
	rawFrame[19] = (tbwFrame.data.timeTag>>32) & 255
	rawFrame[20] = (tbwFrame.data.timeTag>>24) & 255
	rawFrame[21] = (tbwFrame.data.timeTag>>16) & 255
	rawFrame[22] = (tbwFrame.data.timeTag>>8) & 255
	rawFrame[23] = tbwFrame.data.timeTag & 255
	## Data - common
	x = numpy.squeeze(tbwFrame.data.xy[0,:])
	y = numpy.squeeze(tbwFrame.data.xy[1,:])
	## Data - 12 bit
	if tbwFrame.getDataBits() == 12:
		### Round, clip, and convert to unsigned integers
		x = x.round()
		x = x.clip(-2048, 2047)
		x = x.astype(numpy.uint16)
		y = y.round()
		y = y.clip(-2048, 2047)
		y = y.astype(numpy.uint16)
		
		rawFrame[24::3] = (x >> 4) & 255
		rawFrame[25::3] = ((x & 15) << 4) | ((y >> 8) & 15)
		rawFrame[26::3] = y & 255
		
	## Data - 4 bit
	if tbwFrame.getDataBits() == 4:
		### Round, clip, and convert to unsigned integers
		x = x.round()
		x = x.clip(-8, 7)
		x = x.astype(numpy.uint8)
		y = y.round()
		y = y.clip(-8, 7)
		y = y.astype(numpy.uint8)

		rawFrame[24:] = (x << 4) | y
	
	return rawFrame


class SimFrame(tbw.Frame):
	"""tbw.SimFrame extends the lsl.reader.tbw.Frame object to yield a method 
	for easily creating DP ICD-compliant raw TBW frames.  Frames created with
	this method can be written to a file via the methods writeRawFrame() function."""

	def __init__(self, stand=None, frameCount=None, dataBits=12, obsTime=None, xy=None):
		"""Given a list of parameters, build a tbw.SimFrame object.  The parameters
		needed are:
		  + stand id (>0 & <259)
		  + which frame number to create
		  + dataBits (12 or 4)
		  + observation time in seconds since the epoch
		  + 2-D numpy array representing the frame data for both polarizations.
		Not all of these parameters are needed at initialization of the object and
		the values can be added later."""
		
		self.stand = stand
		self.frameCount = frameCount
		self.dataBits = dataBits
		self.obsTime = obsTime
		self.xy = xy
		
		self.header = tbw.FrameHeader()
		self.data = tbw.FrameData()
		
	def __update(self):
		"""Private function to use the object's parameter values to build up 
		a tbw.Frame-like object."""
		
		self.header.frameCount = self.frameCount
		self.header.secondsCount = int(self.obsTime)
		if self.dataBits == 12:
			self.header.tbwID = 32768 | self.stand
		else:
			self.header.tbwID = 32768 | 16384 | self.stand
		
		self.data.timeTag = long(self.obsTime * dp_common.fS)
		self.data.xy = self.xy
		
	def loadFrame(self, tbwFrame):
		"""Populate the a tbw.SimFrame object with a pre-made frame."""
		
		self.header = tbwFrame.header
		self.data = tbwFrame.data
		
		# Back-fill the class' fields to make sure the object is consistent
		## Header
		self.stand = self.header.parseID()
		self.frameCount = self.header.frameCount
		self.dataBits = self.header.getDataBits()
		## Data
		self.obsTime = self.data.timeTag / dp_common.fS
		self.xy = self.data.xy
	
	def isValid(self, raiseErrors=False):
		"""Check if simulated TBW frame is valid or not.  Valid frames return 
		True and invalid frames False.  If the 'raiseErrors' keyword is set, 
		isValid() raises an error when a problem with the frame structure is 
		encountered."""

		# Make sure we have the latest values
		self.__update()

		stand = self.parseID()
		# Is the stand number reasonable?
		if stand == 0 or stand > 258:
			if raiseErrors:
				raise invalidStand()
			return False

		# Is there data loaded into frame.data.xy?
		if self.data.xy is None:
			if raiseErrors:
				raise invalidDataSize()
			return False

		# Does the data shape make sense?
		if self.data.xy.shape[0] != 2:
			if raiseErrors:
				raise invalidDataSize()
			return False

		# Does the data length match the data bits in the header?
		if self.getDataBits() == 12 and self.data.xy.shape[1] != 400:
			if raiseErrors:
				raise invalidDataSize()
			return False
		if self.getDataBits() == 4 and self.data.xy.shape[1] != 1200:
			if raiseErrors:
				raise invalidDataSize()
			return False

		# Does the data type make sense?
		if self.data.xy.dtype.kind == 'c':
			if raiseErrors:
				raise invalidDataType()
			return False

		# If we made it this far, it's valid
		return True

	def createRawFrame(self):
		"""Re-express a simulated TBW frame as a numpy array of unsigned 8-bit 
		integers.  Returns a numpy array if the frame  is valid.  If the frame 
		is not ICD-compliant, a errors.baseSimError-type error is raised."""

		# Make sure we have the latest values
		self.__update()

		self.isValid(raiseErrors=True)
		return frame2frame(self)

	def writeRawFrame(self, fh):
		"""Write a simulated TBW frame to a filehandle if the frame is valid.
		If the frame is not ICD-compliant, a errors.baseSimError-type error is 
		raised."""

		# Make sure we have the latest values
		self.__update()

		rawFrame = self.createRawFrame()
		rawFrame.tofile(fh)

	def __str__(self):
		if self.stand is None:
			return "Empty TBW SimFrame object"
		else:
			return "%i-bit TBW SimFrame for stand %i @ time %i" % (self.dataBits, self.stand, self.obsTime)


# -*- coding: utf-8 -*-

"""Unit test for the lsl.statistics.robust module."""

import os
import time
import unittest
import numpy

from lsl.statistics import robust


__revision__ = "$Revision:1 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"

class robust_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.statistics.robust
	module."""

	def test_mean(self):
		"""Test the outlier-resistant mean function."""

		# Make sure that it can do simple averages
		a = numpy.random.rand(512)
		self.assertAlmostEqual(a.mean(), robust.robustMean(a, Cut=100), 6)

		b = numpy.random.randn(512)
		self.assertAlmostEqual(b.mean(), robust.robustMean(b, Cut=100), 6)

		# Make sure it can reject obvious points
		b = 1.0*a
		b[10] = 1e6
		self.assertTrue(robust.robustMean(b) < b.mean())

	def test_std(self):
		"""Test the outlier-resistant standard deviation function."""

		# Make sure it can reject obvious points
		b = numpy.random.randn(512)
		b[10] = 1e6
		self.assertTrue(robust.robustSigma(b) < b.std())


class robust_test_suite(unittest.TestSuite):
	"""A unittest.TestSuite class which contains all of the lsl.statistics.robust 
	units tests."""
	
	def __init__(self):
		unittest.TestSuite.__init__(self)
		
		loader = unittest.TestLoader()
		self.addTests(loader.loadTestsFromTestCase(robust_tests)) 


if __name__ == '__main__':
	unittest.main()



# -*- coding: utf-8 -*-

"""Unit test for the lsl.correlator.fx module."""

import os
import time
import unittest
import numpy

from lsl.correlator import fx


__revision__ = "$Revision:1 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"

class fx_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.correlator.fx
	module."""

	def setUp(self):
		"""Turn off all numpy warnings."""

		numpy.seterr(all='ignore')

	def test_blackman(self):
		"""Test the Blackman window function."""

		window = {1: [1], 
				2: [-1.3878e-17, -1.3878e-17], 
				4: [-1.3878e-17, 6.3000e-1, 6.3000e-1, -1.3878e-17], 
				8: [-1.3878e-17, 9.0453e-2, 4.5918e-1, 9.2036e-1, 9.2036e-1, 4.5918e-1, 9.0453e-2, -1.3878e-17]}

		for length in window.keys():
			ref = numpy.array(window[length])
			out = fx.blackmanWindow(length)
			
			i = 0
			for r,o in zip(ref, out):
				self.assertAlmostEqual(r, o, 4, "Error on window length %i, value %i" % (length,i))
				i = i + 1

	def test_sinc(self):
		"""Test the sinc window function."""

		window = {1: [1], 
				2: [-3.8980e-17, 1.0], 
				4: [-3.8980e-17, 3.8980e-17, 1.0, 3.8980e-17]}

		for length in window.keys():
			ref = numpy.array(window[length])
			out = fx.sincWindow(length)

			i = 0
			for r,o in zip(ref, out):
				self.assertAlmostEqual(r, o, 5, "Error on window length %i, value %i" % (length,i))
				i = i + 1

	def test_spectra_single(self):
		"""Test the calcSpectra/calcSpectrum functions in single thread mode."""

		fakeData = numpy.random.rand(4,1024) + 3.0
		freq, spectra = fx.calcSpectra(fakeData, DisablePool=True)

	def test_spectra_multi(self):
		"""Test the calcSpectra/calcSpectrum functions in multi-thread mode."""

		fakeData = numpy.random.rand(4,1024) + 3.0
		freq, spectra = fx.calcSpectra(fakeData, DisablePool=False)

	def test_correlator_single(self):
		"""Test the FXCorrelaotr/correlate functions in single thread mode."""

		fakeData = numpy.random.rand(4,1024) + 3.0
		freq, cps = fx.FXCorrelator(fakeData, numpy.array([1,2,3,4]), DisablePool=True)

	def test_correlator_multi(self):
		"""Test the FXCorrelaotr/correlate functions in multi-thread mode."""

		fakeData = numpy.random.rand(4,1024) + 3.0
		freq, cps = fx.FXCorrelator(fakeData, numpy.array([1,2,3,4]), DisablePool=False)



class fx_test_suite(unittest.TestSuite):
	"""A unittest.TestSuite class which contains all of the lsl.correlator.fx
	units tests."""
	
	def __init__(self):
		unittest.TestSuite.__init__(self)
		
		loader = unittest.TestLoader()
		self.addTests(loader.loadTestsFromTestCase(fx_tests)) 


if __name__ == '__main__':
	unittest.main()



# -*- coding: utf-8 -*

import ez_setup
ez_setup.use_setuptools()

import os
import imp
import glob
import unittest

from setuptools import setup, Extension, Distribution, find_packages
try:
	import numpy
except ImportError:
	pass


def get_version():
	"""Read the VERSION file and return the version number as a string."""

	return open('VERSION').read().strip()

def get_description(filename):
	"""Read in a README-type file and return the contents of the DESCRIPTION
	section."""

	desc = ''
	fh = open(filename, 'r')
	lines = fh.readlines()
	fh.close()

	inDescription = False
	for line in lines:
		line = line.replace('\n', '')
		line = line.replace('\t', '')
		if line.find('DESCRIPTION') == 0:
			inDescription = True
			continue
		if line.find('REQUIREMENTS') == 0:
			inDescription = False
			break
		if inDescription:
			desc = ' '.join([desc, line])

	return desc

class LSLDist(Distribution):
	"""Sub-class of setupuptools (distutils.core) Distribution class that fixes
	problems building the libnova extension."""

	def parse_config_files(self, filenames=None):
		# parse cfg file, but remove any pacakge specific options
		# otherwise distutils complains
		
		Distribution.parse_config_files(self, filenames)
		self.get_option_dict('build_ext').pop('libnova_prefix')
		try:
			self.get_option_dict('egg_info').pop('libnova_prefix')
		except:
			pass


setup(
	distclass = LSLDist, 
	name = "lsl", 
	version = get_version(), 
	description = "LWA Software Library", 
	author = "Jayce Dowell", 
	author_email = "jdowell@unm.edu", 
	url = "http://panda.unm.edu/Courses/Dowell/", 
	long_description = get_description('README'), 
	classifiers = ['Development Status :: 4 - Beta',
			'Intended Audience :: Science/Research',
			'Topic :: Scientific/Engineering :: Astronomy'],
	packages = find_packages(), 
	scripts = glob.glob('lsl/scripts/*.py'), 
	setup_requires = ['numpy>=1.2'], 
	install_requires = ['pyfits>=2.1', 'numpy>=1.2', 'scipy>=0.7', 'pyephem>=3.7.3', 'aipy>=0.9.1'], 
	dependency_links = ['http://www.stsci.edu/resources/software_hardware/pyfits'], 
	include_package_data = True, 
	ext_package = 'lsl', 
	ext_modules = [Extension('_libnova', ['lsl/libnova.i']), 
				Extension('astro_array', ['lsl/astro_array.c'], include_dirs=[numpy.get_include()])], 
	test_suite = "lsl.tests.test_lsl.lsl_tests"
) 



# -*- coding: utf-8 -*-

"""Module for seting up DiFX to work with TBW and TBN data.  This includes:
  writing the various configuration files needed by DiFX and
  converting the TBW/TBN data into VDIF format.

This module is under development and is not automatically loaded as part 
of LSL."""

import sys
import numpy
try:
	import cStringIO as StringIO
except ImportError:
	import StringIO

from lsl.misc import geodesy
from lsl import astro as astro
from lsl.common import stations as lwa_common
from lsl.correlator import uvUtils as uvUtils
from lsl.reader import tbw as tbw
from lsl.reader import tbn as tbn
from lsl.reader import errors as errors
from lsl.writer import vdif

__version__ = '0.1'
__revision__ = '$ Revision: 1 $'
__all__ = ['ConfigError', 'DiFXConfig', '__version__', '__revision__', '__all__']

class ConfigError(Exception):
	"""Base exception class for dealing with errors in creating the DiFX 
	configuration files."""

	def __init__(self,  strerror, errno='-1'):
		self.errno = errno
		self.strerror = strerror
		self.args = (errno, strerror)

	def __str__(self):
		return "%s" % self.strerror


class DiFXConfig(object):
	"""Object for creating DiFX input, calc, flag, and joblist configuration 
	files.  The class is initialized with a set of common observation parameters
	and the output files are written using the class's 'write' function.  By 
	default, the outfile files are named:
	  lwa_<mode>_<mjd>.joblist
	  lwa_<mode>_<mjd>_1.input
	  lwa_<mode>_<mjd>_1.calc
	  lwa_<mode>_<mjd>_1.flag

	Once the various configuration parameters have been defined, the config.
	files can be written to disk using DiFXConfig.write() or DiFXConfig.make().
	The difference between the two function is the `write' only writes the 
	config. files while `make' writes the config. files and converts the TBW/
	TBN data into appropriately-named VDIF files."""

	def __init__(self, site=lwa_common.lwa1(), stands=[], mode='TBW', refTime=0.0, intTime=0, nInts=1, bits=12, filter=7, centralFreq=49.0e6):
		"""Initialize all of the values needed to create a complete set of DiFX 
		configuration files.  Required parameters are:
		  + LWA station object, 
		  + list of stands, 
		  + observations mode (TBW or TBN), 
		  + time of first observation, 
		  + integration time in seconds, and
		  + number of integration times to compute.
		For TBW data, the additional information is required:
		  + number of bits in the data (12 or 4).
		For TBN data, the additional information is required:
		  + TBN filter code and 
		  + central frequency in Hz.
		"""

		# Basic info
		self.site = site
		self.stands = stands
		self.mode = mode.upper()
		self.refTime = self.parseRefTime(refTime)
		self.intTime = intTime
		self.nInts = 1

		# Make sure we have a valid mode
		if self.mode not in ['TBW', 'TBN']:
			raise ConfigError("Unknown LWA observing mode '%s' " % self.mode)

		# Default file basename creation
		refMJD = refJD - astro.MJD_OFFSET
		self.basename = "lwa_%s_%i" % (self.mode.lower(), int(refMJD))

		# Mode-specific frequency, bandwidth, and data setup
		if self.mode == 'TBW':
			self.centralFreq = 49.0e6
			self.bandwidth = 98e6
			self.bits = bits
		else:
			self.centralFreq = centralFreq
			self.bandwidth = tbn.filterCodes[filter]
			self.bits = 8

	def setBasename(self, basename):
		"""Override the default file basename of "lwa_<mode>_<MJD>" with a user-
		specified string."""

		self.basename = basename

	def parseRefTime(self, refTime):
		"""Given a time as either a integer, float, string, or datetime object, 
		convert it to a string in the formation 'YYYY-MM-DDTHH:MM:SS'."""

		# Valid time string (modulo the 'T')
		timeRE = re.compile(r'\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}(\.\d+)?')

		if type(refTime).__name__ in ['int', 'long', 'float']:
			refDateTime = datetime.utcfromtimestamp(refTime)
			refTime = refDateTime.strftime("%Y-%m-%dT%H:%M:%S")
		elif type(refTime).__name__ in ['datetime']:
			refTime = refTime.strftime("%Y-%m-%dT%H:%M:%S")
		elif type(refTime).__name__ in ['str']:
			# Make sure that the string times are of the correct format
			if re.match(timeRE, refTime) is None:
				raise RuntimeError("Malformed date/time provided: %s" % refTime)
			else:
				refTime = refTime.replace(' ', 'T', 1)
		else:
			raise RuntimeError("Unknown time format provided.")

		return refTime

	def refTime2AstroDate(self):
		"""Convert a reference time string to an astro.date object."""

		dateStr = self.refTime.replace('T', '-').replace(':', '-').split('-')
		return astro.date(int(dateStr[0]), int(dateStr[1]), int(dateStr[2]), int(dateStr[3]), int(dateStr[4]), float(dateStr[5]))

	def getJoblist(self):
		"""Create a string that contains the list of DiFX jobs that needed to be 
		run."""

		refDate = self.refTime2AstroDate()
		refJD = refDate.to_jd()
		refMJD = refJD - astro.MJD_OFFSET
		stpMJD = refMJD + (self.intTime*self.nInts/astro.SECS_IN_DAY)

		fh = StringIO.StringIO()
		fh.write('exper=%s  v2d=lwa_%s.v2d  pass=%s  mjd=%.7f DiFX=DiFX-1.5.4  vex2difx=1.0.4\n' % (self.mode, self.mode.lower(), self.mode, refMJD))
		fh.write('%s_1 %.7f %.7f 3 6.053 10.91  # %s\n' % (self.basename, refMJD, stpMJD, ' '.join(['LWA%03i' % i for i in self.stands])))
	
		joblistString = fh.getvalue()
		fh.close()

		return joblistString

	def getInput(self):
		"""Create a string that contains a DiFX input file."""

		refDate = self.refTime2AstroDate()
		refJD = refDate.to_jd()
		refMJD = refJD - astro.MJD_OFFSET
		stpMJD = refMJD + (self.intTime*self.nInts/astro.SECS_IN_DAY)

		fh = StringIO.String()
		fh.write('# COMMON SETTINGS ##!\n')
		fh.write('DELAY FILENAME:     %s_1.delay\n' % self.basename)
		fh.write('UVW FILENAME:       %s_1.uvw\n' % self.basename)
		fh.write('CORE CONF FILENAME: %s_1.threads\n' % self.basename)
		fh.write('EXECUTE TIME (SEC): 60\n')
		fh.write('START MJD:          %i\n' % int(refMJD))
		fh.write('START SECONDS:      %i\n' % int((refMJD-int(refMJD))*astro.SECS_IN_DAY))
		fh.write('ACTIVE DATASTREAMS: %i\n' % len(self.stands))
		fh.write('ACTIVE BASELINES:   %i\n' % len(self.stands))
		fh.write('VIS BUFFER LENGTH:  %i\n' % 32)
		fh.write('OUTPUT FORMAT:      SWIN\n')
		fh.write('OUTPUT FILENAME:    %s.difx\n' % self.basename)
		fh.write(' \n')
	
		fh.write('# CONFIGURATIONS ###!\n')
		fh.write('NUM CONFIGURATIONS: 1\n')
		fh.write('CONFIG SOURCE:      %s_default\n' % self.mode)
		fh.write('INT TIME (SEC):     %.4f\n' % self.intTime)
		fh.write('NUM CHANNELS:       512\n')
		fh.write('CHANNELS TO AVERAGE:1\n')
		fh.write('OVERSAMPLE FACTOR:  1\n')
		fh.write('DECIMATION FACTOR:  1\n')
		fh.write('BLOCKS PER SEND:    3125\n')
		fh.write('GUARD BLOCKS:       1\n')
		fh.write('POST-F FRINGE ROT:  FALSE\n')
		fh.write('QUAD DELAY INTERP:  TRUE\n')
		fh.write('WRITE AUTOCORRS:    TRUE\n')
		fh.write('PULSAR BINNING:     FALSE\n')

		for i in list(range(len(self.stands))):
			fh.write('DATASTREAM %i INDEX: %i\n' % (i, i))
		for i in list(range((len(self.stands)*(len(self.stands)+1)/2))):
			fh.write('BASELINE %i INDEX:   %i\n' % (i, i))
		fh.write(' \n')
	
		fh.write('# FREQ TABLE #######!\n')
		fh.write('FREQ ENTRIES:       1\n')
		fh.write('FREQ (MHZ) 0:       %.5f\n' % (self.centralFreq-self.bandwidth/2.0))
		fh.write('BW (MHZ) 0:         %.5f\n' % self.bandwidth)
		fh.write('SIDEBAND 0:         U\n')
		fh.write(' \n')
	
		fh.write('# TELESCOPE TABLE ##!\n')
		fh.write('TELESCOPE ENTRIES:  %i\n' % len(stands))
		i = 0
		for stand in self.stands:
			fh.write('TELESCOPE NAME %i:   LWA%03i\n' % (i, stand))
			fh.write('CLOCK DELAY (us) %i: %.9f\n' % (i, uvUtils.cableDelay(stand, freq=self.centralFreq)*1e6))
			fh.write('CLOCK RATE(us/s) %i: %.9f\n' % (i, 0.0*1e6))
			i = i + 1
		fh.write(' \n')
	
		fh.write('# DATASTREAM TABLE #!\n')
		fh.write('DATASTREAM ENTRIES: %i\n' % len(stands))
		fh.write('DATA BUFFER FACTOR: 32\n')
		fh.write('NUM DATA SEGMENTS:  %i\n' % len(stands))
		i = 0
		for stand in self.stands:
			fh.write('TELESCOPE INDEX:    %i\n' % i)
			fh.write('TSYS:               %.4f\n' % 0.0)
			fh.write('DATA FORMAT:        VDIF\n')
			fh.write('QUANTISATION BITS:  %i\n' % self.bits)
			fh.write('DATA FRAME SIZE:    40004096\n')
			fh.write('DATA SOURCE:        FILE\n')
			fh.write('FILTERBANK USED:    FALSE\n')
			fh.write('NUM FREQS:          1\n')
			fh.write('FREQ TABLE INDEX 0: 0\n')
			fh.write('CLK OFFSET 0 (us):  0.000000\n')
			fh.write('NUM POLS 0:         1\n')
			fh.write('INPUT BAND 0 POL:   X\n')
			fh.write('INPUT BAND 0 INDEX: 0\n')
			i = i + 1
		fh.write(' \n')
	
		fh.write('# BASELINE TABLE ###!\n')
		fh.write('BASELINE ENTRIES:   %i\n' % (len(stands)*(len(stands)+1)/2))
		c = 0
		for i in list(range((len(stands)-1))):
			for j in list(range(1,len(stands))):
				if i == j:
					continue
				fh.write('D/STREAM A INDEX %i: %i\n' % (c, i))
				fh.write('D/STREAM B INDEX %i: %i\n' % (c, j))
				fh.write('NUM FREQS %i:        1\n' % c)
				fh.write('POL PRODUCTS %i/0:   1\n' % c)
				for k in list(range(len(stands))):
					fh.write('D/STREAM A BAND %i: 0\n' % k)
					fh.write('D/STREAM B BAND %i: 0\n' % k)
				c = c + 1
		fh.write(' \n')
	
		fh.write('# DATA TABLE #######!\n')
		i = 0
		for stand in self.stands:
			fh.write('D/STREAM %i FILES:   1\n' % i)
			fh.write('FILE %i/0:           %s-%i.vdif\n' % (i, self.mode, stand))
			i = i + 1
		fh.write(' \n')

		inputString = fh.getvalue()
		fh.close()

		return inputString

	def getCalc(self):
		"""Create a string that contains the bulk of the DiFX calc file."""

		refDate = self.refTime2AstroDate()
		refJD = refDate.to_jd()
		refMJD = refJD - astro.MJD_OFFSET
		stpMJD = refMJD + (self.intTime*self.nInts/astro.SECS_IN_DAY)
		xyzs = uvUtils.getXYZ(self.stands)
		
		fh = StringIO.StringIO()
		fh.write('JOB ID:             %i\n' % 1)
		fh.write('JOB START TIME:     %.7f\n' % refMJD)
		fh.write('JOB STOP TIME:      %.7f\n' % stpMJD)
		fh.write('DUTY CYCLE:         1.000\n')
		fh.write('OBSCODE:            %s\n' % self.mode)
		fh.write('DIFX VERSION:       DiFX-1.5.4\n')
		fh.write('SUBJOB ID:          0\n')
		fh.write('SUBARRAY ID:        0\n')
		fh.write('START MJD:          %.7f\n' % refMJD)
		fh.write('START YEAR:         %i\n' % refDate.years)
		fh.write('START MONTH:        %i\n' % refDate.months)
		fh.write('START DAY:          %i\n' % refDat.days)
		fh.write('START HOUR:         %i\n' % refDat.hours)
		fh.write('START MINUTE:       %i\n' % refDat.minutes)
		fh.write('START SECOND:       %i\n' % int(refDat.seconds))
		fh.write('INCREMENT (SECS):   %.f\n' % self.intTIme)
		fh.write('SPECTRAL AVG:       1\n')
		fh.write('TAPER FUNCTION:     UNIFORM\n')
		fh.write('NUM TELESCOPES:     %i\n' % len(self.stands))
		
		i = 0
		for stand in self.stands:
			siteECI = site.getGeocentricLocation()
			standECI = siteECI + numpy.dot(site.getECEFTransform(), xyzs[i,:])
			
			fh.write('TELESCOPE %i NAME:   LWA%03i\n' % (i, stand))
			fh.write('TELESCOPE %i MOUNT:  azel\n' % i)
			fh.write('TELESCOPE %i OFFSET (m):0.000000\n' % i)
			fh.write('TELESCOPE %i X (m):  %.6f\n' % (i, standECI[0]))
			fh.write('TELESCOPE %i Y (m):  %.6f\n' % (i, standECI[1]))
			fh.write('TELESCOPE %i Z (m):  %.6f\n' % (i, standECI[2]))
			fh.write('TELESCOPE %i SHELF:  NONE\n' % i)
			i = i + 1

		fh.write('NUM SCANS:          1\n')
		fh.write('SCAN 0 POINTS:      10\n')
		fh.write('SCAN 0 START PT:    0\n')
		fh.write('SCAN 0 SRC NAME:    %s_default\n' % self.mode)
		fh.write('SCAN 0 REAL NAME:   ZENITH\n')
		fh.write('SCAN 0 SRC RA:      %.9f\n' % 0.0)
		fh.write('SCAN 0 SRC DEC:     %.9f\n' % 0.0)
		fh.write('SCAN 0 CALCODE:     \n')
		fh.write('SCAN 0 QUAL:        0\n')

		refDate = self.refTime2AstroDate()
		refMJD = refDate.to_jd() - astro.MJD_OFFSET
		eops = geodesy.getEOPRange(refMJD, stpMJD)
		if eops is None:
			eops = [geodesy.EOP(mjd=refMJD)]

		fh.write('NUM EOP:            %i\n' % len(eops))
		i = 0
		for eop in eops:
			fh.write('EOP %i TIME (mjd):   %i\n' % (i, int(eop.mjd)))
			fh.write('EOP %i TAI_UTC (sec):%i\n' % (i, astro.leap_secs(refJD)))
			fh.write('EOP %i UT1_UTC (sec):%.6f\n' % (i, eop.utDiff))
			fh.write('EOP %i XPOLE (arcsec):%.6f\n' % (i, eop.x))
			fh.write('EOP %i YPOLE (arcsec):%.6f\n' % (i, eop.y))
			i = i + 1
		fh.write('NUM SPACECRAFT:     0\n')

		fh.write('DELAY FILENAME:     %s_1.delay\n' % self.basename)
		fh.write('UVW FILENAME:       %s_1.uvw\n' % self.basename)
		fh.write('RATE FILENAME:      %s_1.rate\n' % self.basename)
		fh.write('IM_FILENAME:        %s_1.im\n' % self.basename)
		
		calcString = fh.getvalue()
		fh.close()

		return calcString

	def getFlag(self):
		"""Create a string that contains the flag file associated with this DiFX 
		configuration."""

		fh = StringIO.StringIO()
		fh.write('0')

		flagString = fh.getvalue()
		fh.close()
	
		return flagString

	def write(self):
		inputFile = "%s_1.input" % self.basename
		inputCont = self.getInput()
		
		calcFile = "%s_1.calc" % self.basename
		calcCont = self.getCalc()
		
		flagFile = "%s_1.flag" % self.basename
		flagCont = self.getFlag()

		jobsFile = "%s.joblist" % self.basename
		jobsCont = self.getJoblist()

		names = [inputFile, calcFile, flagFile, jobsFile]
		conts = [inputCont, calcCont, flagCont, jobsCont]
		for name,cont = zip(names, conts):
			print "Writing %i lines to file '%s'" % (len(cont.split('\n')), name)
			fh = open(name, "w")
			fh.write(cont)
			fh.close()

	def convert(self, inputFile):
		"""Given a TBW/TBN file, build a list of VDIF files that are ready to go 
		into DiFX.  This function splits the input DP file into Nstands VDIF files 
		that are named in the format speficied by the configuration writer, i.e.:
		  <mode>-<stand>.vdif
		"""

		if self.mode == 'TBW':
			print "Converting TBW file '%s' to %i VDIF files" % (inputFile, len(self.stands))
			self.__convertTBW(inputFile)
		else:
			print "Converting TBN file '%s' to %i VDIF files" % (inputFile, len(self.stands))
			self.__convertTBN(inputFile)
		
	def __convertTBW(self, inputFile):
		"""Private function call by convert for working with TBW data."""

		fh = open(inputFile, 'rb')
	
		# Make sure that the data is TBW and determine the data length
		test = tbw.readFrame(fh)
		if not test.header.isTBW():
			raise errors.notTBWError()
		fh.seek(0)

		# Create the output files
		ofh = []
		outputFiles = []
		for stand in self.stands:
			outputFiles.append('%s-%i.vdif' % (self.mode, stand))
			ofh.append(open(outputFiles[-1], 'wb'))

		# Read in the data and add it to the VDIF files created above
		for i in range(300000):
			# Read in the next frame and anticipate any problems that could occur
			try:
				cFrame = tbw.readFrame(fh, Verbose=False)
			except errors.eofError:
				break
			except errors.syncError:
				syncCount = syncCount + 1
				continue
			except errors.numpyError:
				break

			stand = cFrame.header.parseID()
			if cFrame.header.frameCount % 10000 == 0:
				print("%2i  %14i  %6.3f  %5i  %5i" % (stand, cFrame.data.timeTag, cFrame.getTime(), cFrame.header.frameCount, cFrame.header.secondsCount))
			
			frame1 = Frame(stand=2*(stand-1)+1, time=cFrame.getTime(), bits=cFrame.getDataBits(), 
							sampleRate = dp_common.fS, data=cFrame.data.xy[0,:])
			frame2 = Frame(stand=2*(stand-1)+2, time=cFrame.getTime(), bits=cFrame.getDataBits(), 
							sampleRate = dp_common.fS, data=cFrame.data.xy[1,:])
			
			frame1.writeRawFrame(ofh[2*(stand-1)+0])
			frame2.writeRawFrame(ofh[2*(stand-1)+1])
		fh.close()

		# Close all of the constituent VDIF files.
		for fh in ofh:
			fh.close()

	def __convertTBN(self, inputFile):
		"""Private function call by convert for working with TBN data."""
	
		fh = open(inputFile, "rb")
		
		test = tbn.readFrame(fh)
		if not test.header.isTBN():
			raise errors.notTBNError()
		fh.seek(0)

		nFpO = tbn.getFramesPerObs(fh)
		nFpO = nFpO[0] + nFpO[1]

		sampleRate = tbn.getSampleRate(fh, nFrames=nFpO)
		nSamples = 340000

		blocks = []
		count = {}
		syncCount = 0
		masterCount = 0
		for i in range(nSamples):
			try:
				cFrame = tbn.readBlock(fh, nFrames=nFpO, SampleRate=sampleRate)
			except errors.eofError:
				break
			except errors.syncError:
				syncCount = syncCount + 1
				continue
			except errors.numpyError:
				break
			blocks.append(cFrame)

			for frame in blocks[-1].x:
				if frame is None:
					print "sync error"
					continue
				stand, pol = frame.parseID()
				if stand not in count.keys():
					count[stand] = 0
				count[stand] = count[stand] + 1

	def make(self, inputFile):
		"""Convience wrapper for write and convert so that one call will create
		everything needed to run DiFX on LWA data."""

		self.write()
		self.convert(inputFile)



# -*- coding: utf-8 -*-

"""Unit test for the lsl.sim.s60/tbw/tbn/drx modules."""

import os
import copy
import unittest
import tempfile
import numpy

from lsl.common.paths import dataBuild as dataPath
from lsl.reader import s60 as s60Reader
from lsl.sim import s60 as s60Writer
from lsl.reader import tbw as tbwReader
from lsl.sim import tbw as tbwWriter
from lsl.reader import tbn as tbnReader
from lsl.sim import tbn as tbnWriter
from lsl.sim import errors


__revision__ = "$Revision:1 $"
__version__  = "0.1"
__author__    = "Jayce Dowell"


s60File = os.path.join(dataPath, 'tests', 's60-test.dat')
tbwFile = os.path.join(dataPath, 'tests', 'tbw-test.dat')
tbnFile = os.path.join(dataPath, 'tests', 'tbn-test.dat')


class fake_S60_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.sim.s60
	module."""

	def test_write_frame(self):
		"""Test that the S60 data writer works for frames."""

		testPath = tempfile.mkdtemp(prefix='fakeS60-', suffix='.tmp')
		testFile = os.path.join(testPath, 's60-test.dat')

		# Read in a S60 frame from the test file
		fh = open(s60File, 'rb')
		origData = s60Reader.readFrame(fh)
		fh.close()
		# Write the data to a S60 test frame
		fh = open(testFile, 'wb')
		rawFrame = s60Writer.frame2frame(origData)
		rawFrame.tofile(fh)
		fh.close()
		# Read in the 
		fh = open(testFile, 'rb')
		fakeData = s60Reader.readFrame(fh)
		fh.close()

		# Test the array size and values
		self.assertEqual(len(fakeData), s60Reader.FrameSize/2)
		for i in range(s60Reader.FrameSize/2):
			self.assertAlmostEqual(fakeData[i].real, origData[i].real, 4)
			self.assertAlmostEqual(fakeData[i].imag, origData[i].imag, 4)

		os.unlink(testFile)
		os.rmdir(testPath)

	def test_write_chunk(self):
		"""Test that the S60 data writer works for chunks."""

		testPath = tempfile.mkdtemp(prefix='fakeS60-', suffix='.tmp')
		testFile = os.path.join(testPath, 's60-test.dat')

		# Read in a S60 frame from the test file
		fh = open(s60File, 'rb')
		origData = s60Reader.readChunk(fh, Chunk=2202)
		fh.close()
		
		# Write the data to a S60 test frame
		fh = open(testFile, 'wb')
		rawFrame = s60Writer.chunk2frame(origData)
		rawFrame.tofile(fh)
		fh.close()
		
		# Read in the 
		fh = open(testFile, 'rb')
		fakeData = s60Reader.readChunk(fh, Chunk=2202)
		fh.close()

		# Test the array size and values
		self.assertEqual(len(fakeData), len(origData))
		for i in range(2202):
			self.assertAlmostEqual(fakeData[i].real, origData[i].real, 4)
			self.assertAlmostEqual(fakeData[i].imag, origData[i].imag, 4)

		os.unlink(testFile)
		os.rmdir(testPath)

	def test_frame_data_errors(self):
		"""Test the different error scenarios in s60.frame2frame."""
		
		# Read in a S60 frame from the test file
		fh = open(s60File, 'rb')
		origData = s60Reader.readChunk(fh, Chunk=2202)
		fh.close()
		
		# Try to write a chunk using frame2frame
		self.assertRaises(errors.invalidDataSize, s60Writer.frame2frame, origData)
		
		# Try to write using only real data with frame2frame
		origData = origData.real
		self.assertRaises(errors.invalidDataType, s60Writer.frame2frame, origData)
	
	def test_chunk_data_errors(self):
		"""Test the different error scenarios in s60.chunk2frame."""

		# Read in a S60 frame from the test file
		fh = open(s60File, 'rb')
		origData = s60Reader.readChunk(fh, Chunk=2202)
		fh.close()
		
		# Try to write using only real data with chunk2frame
		origData = origData.real
		self.assertRaises(errors.invalidDataType, s60Writer.chunk2frame, origData)


class fake_TBW_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.sim.tbw
	module."""

	def test_sim_frame(self):
		"""Test the tbw.SimFrame class."""
		
		# Read in a TBW frame from the test file
		fh = open(tbwFile, 'rb')
		origFrame = tbwReader.readFrame(fh)
		fh.close()
		
		fakeFrame = tbwWriter.SimFrame()
		fakeFrame.loadFrame(origFrame)
		# Test the validity of the SimFrame
		self.assertTrue(fakeFrame.isValid())

	def test_write_frame(self):
		"""Test that the TBW data writer works."""

		testPath = tempfile.mkdtemp(prefix='fakeTBW-', suffix='.tmp')
		testFile = os.path.join(testPath, 'tbw-test.dat')

		# Read in a TBW frame from the test file
		fh = open(tbwFile, 'rb')
		origFrame = tbwReader.readFrame(fh)
		fh.close()
		
		# Write the data to a TBW test frame
		fh = open(testFile, 'wb')
		rawFrame = tbwWriter.frame2frame(origFrame)
		rawFrame.tofile(fh)
		fh.close()
		
		# Read in the 
		fh = open(testFile, 'rb')
		fakeFrame = tbwReader.readFrame(fh)
		fh.close()

		# Test values returned by info functions
		self.assertEqual(fakeFrame.parseID(), origFrame.parseID())
		self.assertEqual(fakeFrame.getDataBits(), origFrame.getDataBits())
		
		# Test raw header values
		self.assertTrue(fakeFrame.header.isTBW())
		self.assertEqual(fakeFrame.header.frameCount, origFrame.header.frameCount)
		self.assertEqual(fakeFrame.header.secondsCount, origFrame.header.secondsCount)
		
		# Test raw data values
		self.assertEqual(fakeFrame.data.timeTag, origFrame.data.timeTag)
		for i in range(400):
			self.assertEqual(fakeFrame.data.xy[0,i], origFrame.data.xy[0,i])
			self.assertEqual(fakeFrame.data.xy[1,i], origFrame.data.xy[1,i])

		os.unlink(testFile)
		os.rmdir(testPath)

	def test_frame_data_errors(self):
		"""Test the data error scenarios when validating a TBW SimFrame ."""
		
		# Read in a TBW frame from the test file
		fh = open(tbwFile, 'rb')
		origFrame = tbwReader.readFrame(fh)
		fh.close()
		
		# Try to validate frame with the wrong data type
		fakeFrame = tbwWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.xy = fakeFrame.data.xy.astype(numpy.complex64)
		self.assertRaises(errors.invalidDataType, fakeFrame.isValid, raiseErrors=True)
		
		# Try to validate frame with the wrong data size
		fakeFrame = tbwWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.xy = None
		self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
		fakeFrame = tbwWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.xy = fakeFrame.data.xy[0,:]
		self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
		fakeFrame = tbwWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.xy = fakeFrame.data.xy[:,0:50]
		self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
		fakeFrame = tbwWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.dataBits = 4
		self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
		
	def test_frame_header_errors(self):
		"""Test the header error scenarios when validating a TBW SimFrame."""
		
		# Read in a TBW frame from the test file
		fh = open(tbwFile, 'rb')
		origFrame = tbwReader.readFrame(fh)
		fh.close()
		
		# Try to validate frame with the wrong stand number
		fakeFrame = tbwWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.stand = 300
		self.assertRaises(errors.invalidStand, fakeFrame.isValid, raiseErrors=True)


class fake_TBN_tests(unittest.TestCase):
	"""A unittest.TestCase collection of unit tests for the lsl.sim.tbn
	module."""

	def test_sim_frame(self):
		"""Test the tbn.SimFrame class."""
		
		# Read in a TBN frame from the test file
		fh = open(tbnFile, 'rb')
		origFrame = tbnReader.readFrame(fh)
		fh.close()
		
		fakeFrame = tbnWriter.SimFrame()
		fakeFrame.loadFrame(origFrame)
		# Test the validity of the SimFrame
		self.assertTrue(fakeFrame.isValid())

	def test_write_frame(self):
		"""Test that the TBN data writer works."""
		
		testPath = tempfile.mkdtemp(prefix='fakeTBN-', suffix='.tmp')
		testFile = os.path.join(testPath, 'tbn-test.dat')

		# Read in a TBN frame from the test file
		fh = open(tbnFile, 'rb')
		origFrame = tbnReader.readFrame(fh)
		fh.close()
		
		# Write the data to a TBN test frame
		fh = open(testFile, 'wb')
		rawFrame = tbnWriter.frame2frame(origFrame)
		rawFrame.tofile(fh)
		fh.close()
		
		# Read in the 
		fh = open(testFile, 'rb')
		fakeFrame = tbnReader.readFrame(fh)
		fh.close()

		# Test values returned by info functions
		self.assertEqual(fakeFrame.parseID()[0], origFrame.parseID()[0])
		self.assertEqual(fakeFrame.parseID()[1], origFrame.parseID()[1])
		
		# Test raw header values
		self.assertTrue(fakeFrame.header.isTBN())
		self.assertEqual(fakeFrame.header.frameCount, origFrame.header.frameCount)
		self.assertEqual(fakeFrame.header.secondsCount, origFrame.header.secondsCount)
		
		# Test raw data values
		self.assertEqual(fakeFrame.data.timeTag, origFrame.data.timeTag)
		for i in range(512):
			self.assertEqual(fakeFrame.data.iq[i].real, origFrame.data.iq[i].real)
			self.assertEqual(fakeFrame.data.iq[i].imag, origFrame.data.iq[i].imag)

		os.unlink(testFile)
		os.rmdir(testPath)

	def test_frame_data_errors(self):
		"""Test the data error scenarios when validating a TBN SimFrame."""
		
		# Read in a TBN frame from the test file
		fh = open(tbnFile, 'rb')
		origFrame = tbnReader.readFrame(fh)
		fh.close()
		
		# Try to validate frame with the wrong data type
		fakeFrame = tbnWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.iq = fakeFrame.data.iq.real
		self.assertRaises(errors.invalidDataType, fakeFrame.isValid, raiseErrors=True)
		
		# Try to validate frame with the wrong data size
		fakeFrame = tbnWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.iq = None
		self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
		fakeFrame = tbnWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.iq = fakeFrame.data.iq[0:50]
		self.assertRaises(errors.invalidDataSize, fakeFrame.isValid, raiseErrors=True)
		
	def test_frame_header_errors(self):
		"""Test the header error scenarios when validating a TBN SimFrame."""
		
		# Read in a TBN frame from the test file
		fh = open(tbnFile, 'rb')
		origFrame = tbnReader.readFrame(fh)
		fh.close()
		
		# Try to validate frame with the wrong stand number
		fakeFrame = tbnWriter.SimFrame()
		fakeFrame.loadFrame(copy.deepcopy(origFrame))
		fakeFrame.stand = 300
		self.assertRaises(errors.invalidStand, fakeFrame.isValid, raiseErrors=True)


class fakedata_test_suite(unittest.TestSuite):
	"""A unittest.TestSuite class which contains all of the lsl.reader units 
	tests."""
	
	def __init__(self):
		unittest.TestSuite.__init__(self)
		
		loader = unittest.TestLoader()
		self.addTests(loader.loadTestsFromTestCase(fake_S60_tests))
		self.addTests(loader.loadTestsFromTestCase(fake_TBW_tests))
		self.addTests(loader.loadTestsFromTestCase(fake_TBN_tests))


if __name__ == '__main__':
	unittest.main()



# -*- coding: utf-8 -*-

"""Python module for creating creating and writing simulated S60 frames to 
a file."""

import numpy

from lsl.common import dp as dp_common
from lsl.reader import s60
from errors import *

__version__ = '0.1'
__revision__ = '$ Revision: 2 $'
__all__ = ['frame2frame', 'chunk2frame', '__version__', '__revision__', '__all__']


def frame2frame(s60Data):
	"""Convert a complex numpy array with the same length as a S60 frame to a 
	raw S60 frame."""

	# The raw frame
	rawFrame = numpy.zeros(s60.FrameSize, dtype=numpy.uint8)

	# Part 1: Sanity checks
	## Compared to the DP data formats (TBW, TBN, and DRX), the S60 data is 
	## relatively barebones.  These checks replace the SimFrame.isValid()
	## function in the DP data formats.
	### Does the data type make sense?
	if s60Data.dtype.kind != 'c':
		raise invalidDataType()
	### Does the data length make sense?
	if s60Data.shape[0] != s60.FrameSize/2:
		raise invalidDataSize()
	
	# Part 2: The data
	i = s60Data.real + 128.0
	q = s60Data.imag + 128.0
	## Round, clip, and convert to unsigned numbers
	i = i.round()
	i = i.clip(0, 255)
	i = i.astype(numpy.uint8)
	q = q.round()
	q = q.clip(0, 255)
	q = q.astype(numpy.uint8)

	rawFrame[0::2] = i
	rawFrame[1::2] = q

	return rawFrame


def chunk2frame(s60Data):
	"""Convert a complex numpy array to a raw S60 frame."""

	# The raw frame
	rawFrame = numpy.zeros(2*len(s60Data), dtype=numpy.uint8)

	# Part 1: Sanity checks
	## Compared to the DP data formats (TBW, TBN, and DRX), the S60 data is 
	## relatively barebones.  These checks replace the SimFrame.isValid()
	## function in the DP data formats.
	### Does the data type make sense?
	if s60Data.dtype.kind != 'c':
		raise invalidDataType()
	### Does the data length make sense?
	if s60Data.shape[0] < 1:
		raise invalidDataSize()
	
	# Part 2: The data
	i = s60Data.real + 128.0
	q = s60Data.imag + 128.0
	## Round, clip, and convert to unsigned numbers
	i = i.round()
	i = i.clip(0, 255)
	i = i.astype(numpy.uint8)
	q = q.round()
	q = q.clip(0, 255)
	q = q.astype(numpy.uint8)

	rawFrame[0::2] = i
	rawFrame[1::2] = q

	return rawFrame



# -*- coding: utf-8 -*-

"""
Classes and methods to model sky brightness and visibility.
"""

##############################################################
# $Id: skymap.py 90 2010-05-20 23:17:17Z dwood $
##############################################################

### This module handles the skymaps from the 74 MHz skymap.
### David Munton, ARL:UT Jan 2007

from numpy import *
import os,sys

import pyfits

from lsl import astro_array

__revision__  = "$Revision: 90 $"
__version__   = "dev"
__author__    = "J.York"
__maintainer__ = "Jayce Dowell"


### This code is the base class for the sky map. It takes as input a skymap file name and frequency to
### which the skymap corresponds.  It has the following methods:
###     _init_ - takes the array coordinate filename as an input argument.
###     NormalizePower - Converts the skymap powers (in Kelvin radiated into 4 pi ster) into a power seen at the antenna.
###     ComputeTotalPowerFromSky - Sums the power for all sources in the sky
###     ScaleSourcePowerstoFrequency - Scales the skymap from the base 73.8 MHz to the desired frequency.

class SkyMapError(Exception):
    """Base class for SkyMap exceptions"""
    pass

class SkyMapLoadError(SkyMapError):
    """Exception raised when a SkyMap fails to load"""
    pass

class SkyMapPowerError(SkyMapError):
    """Exception raised when an operation on a normalized SkyMap fails to compute (such as summing or multiplying over an empty data set)"""
    pass

class SkyMap(object):
    "The class for handling the model sky brightness maps."

    # Class data 
    
    degToRad = (pi/180.)             # Usual conversion factor
    
    def __init__(self, skyMapFileName=None, freqMHz=73.9):
        """Initialize the SkyMap object with an optional full file path to the skymap file."""
        if skyMapFileName is None:
            from lsl.common.paths import data as dataPath
            skyMapFileName = os.path.join(dataPath, 'skymap', 'LFmap_73.9.fits')
        
        fits = pyfits.open(skyMapFileName)
        
        hdr = fits[0].header
        
        self.numPixelsX = hdr['NAXIS1']
        self.numPixelsY = hdr['NAXIS2']
        
        self.freq = float(hdr['FINALFRQ'].split()[0])
  
        ra = (arange(1, self.numPixelsX + 1, dtype = float32) - hdr['CRPIX1']) * hdr['CDELT1']
        self.ra = ra.reshape(1, len(ra)).repeat(self.numPixelsY, 0).ravel()
        
        dec = (arange(1, self.numPixelsY + 1, dtype = float32) - hdr['CRPIX2']) * hdr['CDELT2']
        self.dec = dec.repeat(self.numPixelsX, -1)
        
        self._power = empty((self.numPixelsX * self.numPixelsY,), float32) 
        data = fits[0].data
        
        for i in range(self.numPixelsY):
            ii = i * self.numPixelsX
            for j in range(self.numPixelsX):
                self._power[ii + j] = data[i, j]
        
        fits.close()
        
    def NormalizePower(self):
        """Compute the skymap power (total power radiated into 4 pi steradians) into a power at antenna, based on pixel count."""
        return self._power*cos(self.dec * self.degToRad) # The cosine term is the projection of the receiving area onto the direction of the source"

    def ComputeTotalPowerFromSky(self):
        """Compute and return the the total power from the sky."""
        if len(self._power) == 0: raise SkyMapPowerError("self._power contains 0 elements")
        return self.NormalizePower().sum()
        
### This code is the base class for the sky map visible at a specific location. It takes as input a skymap file name and frequency to
### which the skymap corresponds.  It inherits from class SkyMap. It has the following methods:
###     _init_ - takes the array coordinate filename as an input argument.
###     ComputeDirectionCosines - Computes the direction cosines 
###     ComputeTotalPowerFromVisibleSky - Sums the power for all visible sources in the sky

class ProjectedSkyMap(object):
    """The class for handling the model sky brightness maps over a particular site."""

        
    def __init__(self, skyMapObject, lat, lon, utc_jd):
        """Initialize the skymap at input lat,lon (decimal degrees) and time (in UTC julian day)."""
        self.skyMapObject=skyMapObject

        assert -90<=lat<=90, ValueError('lat = %g not in [-90,90]' % lat)
        assert -360<=lon<=360, ValueError('lon = %g not in [-360,360]' % lon)
        self.lat=lat
        self.lon=lon
        self.time=utc_jd

        # Compute the ra and dec locations to a altitude and azimuth. This requires a lat/lon and a time (UTC). Alt and Az are compressed to show only the visible source.
        self.alt,self.az=astro_array.hrz_from_equ(self.lon, self.lat, self.time, self.skyMapObject.ra, self.skyMapObject.dec)  ### Replace me with a dot product instead of astro_array

        # Compress the az/alt so that only the visible sources are available. "
        visibleMask = self.alt > 0.0
        
        #Compress arrays to hide non-visible sources
        self.visibleAlt = compress(visibleMask, self.alt)
        self.visibleAz = compress(visibleMask, self.az)
        self.visiblePower = compress(visibleMask, self.skyMapObject._power)
        pixelSolidAngle = (2.0 * pi)**2/(self.skyMapObject.numPixelsX*self.skyMapObject.numPixelsY)
        fractionSolidAngle = (1.0/4.0*pi) * pixelSolidAngle
        normalizedPower = self.skyMapObject.NormalizePower() * fractionSolidAngle  # The cosine term is the projection of the receiving area onto the direction of the source"
        self.visibleNormalizedPower = compress(visibleMask, normalizedPower)
        self.visibleRa = compress(visibleMask, self.skyMapObject.ra)
        self.visibleDec = compress(visibleMask, self.skyMapObject.dec)

    def ComputeDirectionCosines(self):
        "Compute the direction cosines and return the tuple of arrays(l,m,n)."
        altRad = self.visibleAlt*self.skyMapObject.degToRad
        azRad = self.visibleAz*self.skyMapObject.degToRad
        l = cos(altRad)*sin(azRad)
        m = cos(altRad)*cos(azRad)
        n = sin(altRad)
        return (l,m,n)

    def ComputeTotalPowerFromVisibleSky(self):
        """Compute and return the the total power from visible portion of the sky."""
        if len(self.visibleNormalizedPower) == 0: raise SkyMapPowerError("visibleNormalizedPower contains 0 elements")
        totalVisiblePower = sum(self.visibleNormalizedPower)
        return totalVisiblePower

